# This file describes eFuses for ESP32-H21 chip
#
# SPDX-FileCopyrightText: 2024 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: GPL-2.0-or-later

import struct
import sys
import time

import reedsolo
from bitstring import BitArray

import esptool
from esptool.logger import log

from .. import base_fields
from ..mem_definition_base import Field
from .mem_definition import EfuseDefineBlocks, EfuseDefineFields, EfuseDefineRegisters


class EfuseBlock(base_fields.EfuseBlockBase):
    def len_of_burn_unit(self):
        # The writing register window is 8 registers for any blocks.
        # len in bytes
        return 8 * 4

    def __init__(self, parent, param, skip_read=False):
        parent.read_coding_scheme()
        super().__init__(parent, param, skip_read=skip_read)

    def apply_coding_scheme(self):
        data = self.get_raw(from_read=False)[::-1]
        if len(data) < self.len_of_burn_unit():
            add_empty_bytes = self.len_of_burn_unit() - len(data)
            data = data + (b"\x00" * add_empty_bytes)
        if self.get_coding_scheme() == self.parent.REGS.CODING_SCHEME_RS:
            # takes 32 bytes
            # apply RS encoding
            rs = reedsolo.RSCodec(12)
            # 32 byte of data + 12 bytes RS
            encoded_data = rs.encode([x for x in data])
            words = struct.unpack("<" + "I" * 11, encoded_data)
            # returns 11 words (8 words of data + 3 words of RS coding)
        else:
            # takes 32 bytes
            words = struct.unpack("<" + ("I" * (len(data) // 4)), data)
            # returns 8 words
        return words


class EspEfuses(base_fields.EspEfusesBase):
    """
    Wrapper object to manage the efuse fields in a connected ESP bootloader
    """

    def __init__(
        self,
        esp,
        skip_connect=False,
        debug=False,
        do_not_confirm=False,
        extend_efuse_table=None,
    ):
        super().__init__(esp, skip_connect, debug, do_not_confirm, extend_efuse_table)
        self.Blocks = EfuseDefineBlocks()
        self.Fields = EfuseDefineFields(extend_efuse_table)
        self.REGS = EfuseDefineRegisters
        self.BURN_BLOCK_DATA_NAMES = self.Blocks.get_burn_block_data_names()
        self.BLOCKS_FOR_KEYS = self.Blocks.get_blocks_for_keys()
        if esp.CHIP_NAME != "ESP32-H21":
            raise esptool.FatalError(
                f"Expected the 'esp' param for ESP32-H21 chip but got for '{esp.CHIP_NAME}'."
            )
        if not skip_connect:
            flags = self._esp.get_security_info()["flags"]
            GET_SECURITY_INFO_FLAG_SECURE_DOWNLOAD_ENABLE = 1 << 2
            if flags & GET_SECURITY_INFO_FLAG_SECURE_DOWNLOAD_ENABLE:
                raise esptool.FatalError(
                    "Secure Download Mode is enabled. The tool can not read eFuses."
                )
        self.blocks = [
            EfuseBlock(self, self.Blocks.get(block), skip_read=skip_connect)
            for block in self.Blocks.BLOCKS
        ]
        if not skip_connect:
            self.get_coding_scheme_warnings()
        self.efuses = self._convert_efuse_defs(self.Fields.EFUSES)
        self.efuses += self._convert_efuse_defs(self.Fields.KEYBLOCKS)
        if skip_connect:
            self.efuses += self._convert_efuse_defs(
                self.Fields.BLOCK2_CALIBRATION_EFUSES
            )
        else:
            if self.get_block_version() >= 2:
                self.efuses += self._convert_efuse_defs(
                    self.Fields.BLOCK2_CALIBRATION_EFUSES
                )
            self.efuses += self._convert_efuse_defs(self.Fields.CALC)

    def _convert_efuse_defs(self, efuse_defs):
        return [EfuseField.convert(self, efuse) for efuse in efuse_defs]

    def _get_lazy_efuse_groups(self):
        return [self.Fields.BLOCK2_CALIBRATION_EFUSES]

    def read_coding_scheme(self):
        self.coding_scheme = self.REGS.CODING_SCHEME_RS

    def print_status_regs(self):
        log.print("")
        self.blocks[0].print_block(self.blocks[0].err_bitarray, "err__regs", debug=True)
        log.print(
            "{:27} 0x{:08x}".format(
                "EFUSE_RD_RS_ERR0_REG", self.read_reg(self.REGS.EFUSE_RD_RS_ERR0_REG)
            )
        )
        log.print(
            "{:27} 0x{:08x}".format(
                "EFUSE_RD_RS_ERR1_REG", self.read_reg(self.REGS.EFUSE_RD_RS_ERR1_REG)
            )
        )

    def efuse_controller_setup(self):
        self.set_efuse_timing()
        self.clear_pgm_registers()
        self.wait_efuse_idle()

    def write_efuses(self, block):
        self.efuse_program(block)
        return self.get_coding_scheme_warnings(silent=True)

    def clear_pgm_registers(self):
        self.wait_efuse_idle()
        for r in range(
            self.REGS.EFUSE_PGM_DATA0_REG, self.REGS.EFUSE_PGM_DATA0_REG + 32, 4
        ):
            self.write_reg(r, 0)

    def wait_efuse_idle(self):
        deadline = time.time() + self.REGS.EFUSE_BURN_TIMEOUT
        while time.time() < deadline:
            cmds = self.REGS.EFUSE_PGM_CMD | self.REGS.EFUSE_READ_CMD
            if self.read_reg(self.REGS.EFUSE_CMD_REG) & cmds == 0:
                if self.read_reg(self.REGS.EFUSE_CMD_REG) & cmds == 0:
                    # Due to a hardware error, we have to read READ_CMD again
                    # to make sure the efuse clock is normal.
                    # For PGM_CMD it is not necessary.
                    return
        raise esptool.FatalError(
            "Timed out waiting for eFuse controller command to complete"
        )

    def efuse_program(self, block):
        self.wait_efuse_idle()
        self.write_reg(self.REGS.EFUSE_CONF_REG, self.REGS.EFUSE_WRITE_OP_CODE)
        self.write_reg(self.REGS.EFUSE_CMD_REG, self.REGS.EFUSE_PGM_CMD | (block << 2))
        self.wait_efuse_idle()
        self.clear_pgm_registers()
        self.efuse_read()

    def efuse_read(self):
        self.wait_efuse_idle()
        self.write_reg(self.REGS.EFUSE_CONF_REG, self.REGS.EFUSE_READ_OP_CODE)
        # need to add a delay after triggering EFUSE_READ_CMD, as ROM loader checks some
        # efuse registers after each command is completed
        # if ENABLE_SECURITY_DOWNLOAD or DIS_DOWNLOAD_MODE is enabled by the current cmd, then we need to try to reconnect to the chip.
        try:
            self.write_reg(
                self.REGS.EFUSE_CMD_REG, self.REGS.EFUSE_READ_CMD, delay_after_us=1000
            )
            self.wait_efuse_idle()
        except esptool.FatalError:
            secure_download_mode_before = self._esp.secure_download_mode

            try:
                self._esp = self.reconnect_chip(self._esp)
            except esptool.FatalError:
                log.print("Can not re-connect to the chip.")
                if not self["DIS_DOWNLOAD_MODE"].get() and self[
                    "DIS_DOWNLOAD_MODE"
                ].get(from_read=False):
                    log.print(
                        "This is the correct behavior as we are actually burning "
                        "DIS_DOWNLOAD_MODE which disables the connection to the chip."
                    )
                    log.print("DIS_DOWNLOAD_MODE is enabled.")
                    log.print("Successful.")
                    sys.exit(0)  # finish without errors
                raise

            log.print("Established a connection with the chip.")
            if self._esp.secure_download_mode and not secure_download_mode_before:
                log.print("Secure download mode is enabled.")
                if not self["ENABLE_SECURITY_DOWNLOAD"].get() and self[
                    "ENABLE_SECURITY_DOWNLOAD"
                ].get(from_read=False):
                    log.print(
                        "espefuse can not continue to work in Secure download mode."
                    )
                    log.print("ENABLE_SECURITY_DOWNLOAD is enabled.")
                    log.print("Successful.")
                    sys.exit(0)  # finish without errors
            raise

    def set_efuse_timing(self):
        """Set timing registers for burning efuses"""
        # Configure clock
        apb_freq = self.get_crystal_freq()
        # Based on `CONFIG_SOC_XTAL_SUPPORT_32M=y` for this target from ESP-IDF configuration
        if apb_freq != 32:
            raise esptool.FatalError(
                f"The eFuse supports only xtal=32M (xtal was {apb_freq}"
            )
        # TODO: [ESP32H21] IDF-11506

    def get_coding_scheme_warnings(self, silent=False):
        """Check if the coding scheme has detected any errors."""
        old_addr_reg = 0
        reg_value = 0
        ret_fail = False
        for block in self.blocks:
            if block.id == 0:
                words = [
                    self.read_reg(self.REGS.EFUSE_RD_REPEAT_ERR0_REG + offs * 4)
                    for offs in range(5)
                ]
                block.err_bitarray.pos = 0
                for word in reversed(words):
                    block.err_bitarray.overwrite(BitArray(f"uint:32={word}"))
                block.num_errors = block.err_bitarray.count(True)
                block.fail = block.num_errors != 0
            else:
                addr_reg, err_num_mask, err_num_offs, fail_bit = self.REGS.BLOCK_ERRORS[
                    block.id
                ]
                if err_num_mask is None or err_num_offs is None or fail_bit is None:
                    continue
                if addr_reg != old_addr_reg:
                    old_addr_reg = addr_reg
                    reg_value = self.read_reg(addr_reg)
                block.fail = reg_value & (1 << fail_bit) != 0
                block.num_errors = (reg_value >> err_num_offs) & err_num_mask
            ret_fail |= block.fail
            if not silent and (block.fail or block.num_errors):
                log.print(
                    f"Error(s) in BLOCK{block.id} "
                    f"[ERRORS:{block.num_errors} FAIL:{block.fail}]."
                )
        if (self.debug or ret_fail) and not silent:
            self.print_status_regs()
        return ret_fail

    def summary(self):
        # TODO add support set_flash_voltage - "Flash voltage (VDD_SPI)"
        return ""


class EfuseField(base_fields.EfuseFieldBase):
    @staticmethod
    def convert(parent: base_fields.EspEfusesBase, efuse: Field) -> "EfuseField":
        return {
            "mac": EfuseMacField,
            "keypurpose": EfuseKeyPurposeField,
            "t_sensor": EfuseTempSensor,
            "adc_tp": EfuseAdcPointCalibration,
            "wafer": EfuseWafer,
        }.get(efuse.class_type, EfuseField)(parent, efuse)


class EfuseTempSensor(base_fields.EfuseTempSensor, EfuseField):
    pass


class EfuseAdcPointCalibration(base_fields.EfuseAdcPointCalibration, EfuseField):
    pass


class EfuseMacField(base_fields.EfuseMacFieldBase, EfuseField):
    pass


class EfuseWafer(base_fields.EfuseWaferBase, EfuseField):
    def get(self, from_read=True):
        # TODO: [ESP32H21] IDF-11506
        return 0


class EfuseKeyPurposeField(base_fields.EfuseKeyPurposeFieldBase, EfuseField):
    # fmt: off
    KEY_PURPOSES = [
        ("USER",                         0,  None,       None,      "no_need_rd_protect"),   # User purposes (software-only use)
        ("ECDSA_KEY",                    1,  None,       "Reverse", "need_rd_protect"),      # ECDSA key
        ("RESERVED",                     2,  None,       None,      "no_need_rd_protect"),   # Reserved
        ("XTS_AES_128_KEY",              4,  None,       "Reverse", "need_rd_protect"),      # XTS_AES_128_KEY (flash/PSRAM encryption)
        ("HMAC_DOWN_ALL",                5,  None,       None,      "need_rd_protect"),      # HMAC Downstream mode
        ("HMAC_DOWN_JTAG",               6,  None,       None,      "need_rd_protect"),      # JTAG soft enable key (uses HMAC Downstream mode)
        ("HMAC_DOWN_DIGITAL_SIGNATURE",  7,  None,       None,      "need_rd_protect"),      # Digital Signature peripheral key (uses HMAC Downstream mode)
        ("HMAC_UP",                      8,  None,       None,      "need_rd_protect"),      # HMAC Upstream mode
        ("SECURE_BOOT_DIGEST0",          9,  "DIGEST",   None,      "no_need_rd_protect"),   # SECURE_BOOT_DIGEST0 (Secure Boot key digest)
        ("SECURE_BOOT_DIGEST1",          10, "DIGEST",   None,      "no_need_rd_protect"),   # SECURE_BOOT_DIGEST1 (Secure Boot key digest)
        ("SECURE_BOOT_DIGEST2",          11, "DIGEST",   None,      "no_need_rd_protect"),   # SECURE_BOOT_DIGEST2 (Secure Boot key digest)
    ]
    # fmt: on



# This file describes eFuses controller for ESP32 chip
#
# SPDX-FileCopyrightText: 2020-2026 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: GPL-2.0-or-later

import re
from abc import ABC, abstractmethod
from typing import Any

from bitstring import BitStream

from espefuse.efuse.mem_definition_base import (
    BlockDefinition,
    EfuseBlocksBase,
    EfuseFieldsBase,
    EfuseRegistersBase,
    Field,
)
from esptool.logger import log


class EmulateEfuseControllerBase(ABC):
    """The class for virtual efuse operations. Using for HOST_TEST."""

    CHIP_NAME: str = ""
    mem: BitStream
    debug: bool = False
    Blocks: type[EfuseBlocksBase]
    Fields: EfuseFieldsBase
    REGS: type[EfuseRegistersBase]
    USB_JTAG_SERIAL_PID: int = 0x1001

    def __init__(self, efuse_file: str | None = None, debug: bool = False):
        self.debug = debug
        self.efuse_file = efuse_file
        if self.efuse_file:
            try:
                self.mem = BitStream(
                    bytes=open(self.efuse_file, "rb").read(),
                    length=self.REGS.EFUSE_MEM_SIZE * 8,
                )
            except (ValueError, FileNotFoundError):
                # the file is empty or does not fit the length.
                self.mem = BitStream(length=self.REGS.EFUSE_MEM_SIZE * 8)
                self.mem.set(0)
                self.mem.tofile(open(self.efuse_file, "a+b"))
        else:
            # efuse_file is not provided
            #  it means we do not want to keep the result of efuse operations
            self.mem = BitStream(self.REGS.EFUSE_MEM_SIZE * 8)
            self.mem.set(0)

    """ esptool method start >> """

    @abstractmethod
    def get_major_chip_version(self) -> int:
        pass

    @abstractmethod
    def get_minor_chip_version(self) -> int:
        pass

    def get_chip_description(self) -> str:
        major_rev = self.get_major_chip_version()
        minor_rev = self.get_minor_chip_version()
        return f"{self.CHIP_NAME} (revision v{major_rev}.{minor_rev})"

    def get_chip_revision(self) -> int:
        return self.get_major_chip_version() * 100 + self.get_minor_chip_version()

    def read_efuse(self, n: int, block: int = 0) -> int:
        """Read the nth word of the ESP3x EFUSE region."""
        blk = self.Blocks.get(self.Blocks.BLOCKS[block])
        return self.read_reg(blk.rd_addr + (4 * n))

    def read_reg(self, addr: int) -> int:
        self.mem.pos = self.mem.length - ((addr - self.REGS.DR_REG_EFUSE_BASE) * 8 + 32)
        return int(self.mem.read("uint:32"))

    def write_reg(
        self,
        addr: int,
        value: int,
        mask: int = 0xFFFFFFFF,
        delay_us: int = 0,
        delay_after_us: int = 0,
    ) -> None:
        self.mem.pos = self.mem.length - ((addr - self.REGS.DR_REG_EFUSE_BASE) * 8 + 32)
        self.mem.overwrite("uint:32={}".format(value & mask))
        self.handle_writing_event(addr, value)

    def update_reg(self, addr: int, mask: int, new_val: int) -> None:
        position = self.mem.length - ((addr - self.REGS.DR_REG_EFUSE_BASE) * 8 + 32)
        self.mem.pos = position
        cur_val = self.mem.read("uint:32")
        self.mem.pos = position
        self.mem.overwrite("uint:32={}".format(cur_val | (new_val & mask)))

    def write_efuse(self, n: int, value: int, block: int = 0) -> None:
        """Write the nth word of the ESP3x EFUSE region."""
        blk = self.Blocks.get(self.Blocks.BLOCKS[block])
        self.write_reg(blk.wr_addr + (4 * n), value)

    """ << esptool method end """

    def handle_writing_event(self, addr: int, value: int) -> None:
        self.save_to_file()

    def save_to_file(self) -> None:
        if self.efuse_file:
            with open(self.efuse_file, "wb") as f:
                self.mem.tofile(f)

    def handle_coding_scheme(self, blk: BlockDefinition, data: BitStream) -> BitStream:
        return data

    def copy_blocks_wr_regs_to_rd_regs(self, updated_block: int | None = None) -> None:
        for b in reversed(self.Blocks.BLOCKS):
            blk = self.Blocks.get(b)
            if updated_block is not None:
                if blk.id != updated_block:
                    continue
            data = self.read_block(blk.id, wr_regs=True)
            if self.debug:
                log.print(blk.name, data.hex)
            plain_data = self.handle_coding_scheme(blk, data)
            plain_data = self.check_wr_protection_area(blk.id, plain_data)
            self.update_block(blk, plain_data)

    def clean_blocks_wr_regs(self) -> None:
        for b in self.Blocks.BLOCKS:
            blk = self.Blocks.get(b)
            for offset in range(0, blk.len * 4, 4):
                wr_addr = blk.wr_addr + offset
                self.write_reg(wr_addr, 0)

    def read_field(self, name: str, bitstring: bool = True) -> Any:
        for field in self.Fields.EFUSES:
            if field.name == name:
                block = self.read_block(field.block)
                if field.type.startswith("bool"):
                    field_len = 1
                else:
                    match = re.search(r"\d+", field.type)
                    if match:
                        field_len = int(match.group())
                    else:
                        raise ValueError(f"Invalid field type: {field.type}")
                    if field.type.startswith("bytes"):
                        field_len *= 8
                block.pos = block.length - (field.word * 32 + field.pos + field_len)  # type: ignore
                if bitstring:
                    return block.read(field_len)
                else:
                    return block.read(field.type)
        raise ValueError(f"Field {name} not found")

    def get_bitlen_of_block(self, blk: BlockDefinition, wr: bool = False) -> int:
        return 32 * blk.len

    def read_block(self, idx: int, wr_regs: bool = False) -> BitStream:
        for b in self.Blocks.BLOCKS:
            blk = self.Blocks.get(b)
            if blk.id == idx:
                blk_len_bits = self.get_bitlen_of_block(blk, wr=wr_regs)
                addr = blk.wr_addr if wr_regs else blk.rd_addr
                self.mem.pos = self.mem.length - (
                    (addr - self.REGS.DR_REG_EFUSE_BASE) * 8 + blk_len_bits
                )
                block = self.mem.read(blk_len_bits)
                break
        else:
            raise ValueError(f"Block {idx} not found")
        return block

    def update_block(self, blk: BlockDefinition, wr_data: BitStream) -> None:
        wr_data = self.read_block(blk.id) | wr_data
        self.overwrite_mem_from_block(blk, wr_data)

    def overwrite_mem_from_block(
        self, blk: BlockDefinition, wr_data: BitStream
    ) -> None:
        self.mem.pos = self.mem.length - (
            (blk.rd_addr - self.REGS.DR_REG_EFUSE_BASE) * 8 + wr_data.len
        )
        self.mem.overwrite(wr_data)

    def check_wr_protection_area(self, num_blk: int, wr_data: BitStream) -> BitStream:
        # checks fields which have the write protection bit.
        # if the write protection bit is set, we need to protect that area from changes.
        write_disable_bit = self.read_field("WR_DIS", bitstring=False)
        mask_wr_data = BitStream(len(wr_data))
        mask_wr_data.set(0)
        blk = self.Blocks.get(self.Blocks.BLOCKS[num_blk])
        if blk.write_disable_bit is not None and write_disable_bit & (
            1 << blk.write_disable_bit
        ):
            mask_wr_data.set(1)
        else:
            for field in self.Fields.EFUSES:
                if blk.id == field.block and field.block == num_blk:
                    if field.write_disable_bit is not None and write_disable_bit & (
                        1 << field.write_disable_bit
                    ):
                        data: BitStream = self.read_field(field.name)
                        data.set(1)
                        mask_wr_data.pos = mask_wr_data.length - (
                            field.word * 32 + field.pos + data.len  # type: ignore
                        )
                        mask_wr_data.overwrite(data)
        mask_wr_data.invert()
        return wr_data & mask_wr_data

    @staticmethod
    def get_read_disable_mask(blk: BlockDefinition | Field) -> int:
        mask = 0
        if isinstance(blk.read_disable_bit, list):
            for i in blk.read_disable_bit:
                mask |= 1 << i
        elif blk.read_disable_bit is not None:
            mask = 1 << blk.read_disable_bit
        return mask

    def set_read_disable_bits(
        self,
        block: BitStream,
        read_disable_bit: int,
        blk: BlockDefinition,
    ) -> None:
        block.set(0)

    def check_rd_protection_area(self) -> None:
        # checks fields which have the read protection bits.
        # if the read protection bit is set then we need to reset this field to 0.
        read_disable_bit = self.read_field("RD_DIS", bitstring=False)
        for b in self.Blocks.BLOCKS:
            blk = self.Blocks.get(b)
            block = self.read_block(blk.id)
            if (
                blk.read_disable_bit is not None
                and read_disable_bit & self.get_read_disable_mask(blk)
            ):
                self.set_read_disable_bits(block, read_disable_bit, blk)
            else:
                for field in self.Fields.EFUSES:
                    if (
                        blk.id == field.block
                        and field.read_disable_bit is not None
                        and read_disable_bit & self.get_read_disable_mask(field)
                    ):
                        raw_data: BitStream = self.read_field(field.name)
                        raw_data.set(0)
                        block.pos = block.length - (
                            field.word * 32 + field.pos + raw_data.length  # type: ignore
                        )
                        block.overwrite(BitStream(raw_data.length))
            self.overwrite_mem_from_block(blk, block)

    def clean_mem(self) -> None:
        self.mem.set(0)
        if self.efuse_file:
            with open(self.efuse_file, "wb") as f:
                self.mem.tofile(f)



# This file helps to parse CSV eFuse tables
#
# SPDX-FileCopyrightText: 2024 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import re
import sys

from esptool.logger import log


class CSVFuseTable(list):
    @classmethod
    def from_csv(cls, csv_contents):
        res = CSVFuseTable()
        lines = csv_contents.splitlines()

        def expand_vars(f):
            f = os.path.expandvars(f)
            m = re.match(r"(?<!\\)\$([A-Za-z_]\w*)", f)
            if m:
                raise InputError(f"unknown variable '{m.group(1)}'")
            return f

        for line_no, line in enumerate(lines):
            line = expand_vars(line).strip()
            if line.startswith("#") or len(line) == 0:
                continue
            try:
                res.append(FuseDefinition.from_csv(line))
            except InputError as err:
                raise InputError(f"Error at line {line_no + 1}: {err}")
            except Exception:
                sys.stderr.write(f"Unexpected error parsing line {line_no + 1}: {line}")
                raise

        # fix up missing bit_start
        last_efuse_block = None
        for i in res:
            if last_efuse_block != i.efuse_block:
                last_end = 0
            if i.bit_start is None:
                i.bit_start = last_end
            last_end = i.bit_start + i.bit_count
            last_efuse_block = i.efuse_block

        res.verify_duplicate_name()

        # fix up missing field_name
        last_field = None
        for i in res:
            if i.field_name == "":
                if last_field is None:
                    raise InputError(
                        f"Error at line {line_no + 1}: {i} missing field name"
                    )
                elif last_field is not None:
                    i.field_name = last_field.field_name
            last_field = i

        # fill group
        names = [p.field_name for p in res]
        duplicates = set(n for n in names if names.count(n) > 1)
        for dname in duplicates:
            i_count = 0
            for p in res:
                if p.field_name != dname:
                    continue
                if len(duplicates.intersection([p.field_name])) != 0:
                    p.field_name = f"{p.field_name}_{i_count}"
                    if p.alt_names:
                        p.alt_names = f"{p.alt_names}_{i_count}"
                    i_count += 1
                else:
                    i_count = 0

        for p in res:
            p.field_name = p.field_name.replace(".", "_")
            if p.alt_names:
                p.alt_names = p.alt_names.replace(".", "_")
        res.verify_duplicate_name()
        return res

    def verify_duplicate_name(self):
        # check on duplicate name
        names = [p.field_name for p in self]
        names += [name.replace(".", "_") for name in names if "." in name]
        duplicates = set(n for n in names if names.count(n) > 1)

        # print sorted duplicate partitions by name
        if len(duplicates) != 0:
            fl_error = False
            for p in self:
                field_name = p.field_name + p.group
                if field_name != "" and len(duplicates.intersection([field_name])) != 0:
                    fl_error = True
                    log.print(
                        f"Field at {p.field_name}, {p.efuse_block}, "
                        f"{p.bit_start}, {p.bit_count} have duplicate field_name"
                    )
            if fl_error is True:
                raise InputError("Field names must be unique")

    def check_struct_field_name(self):
        # check that structured fields have a root field
        for p in self:
            if "." in p.field_name:
                name = ""
                for sub in p.field_name.split(".")[:-1]:
                    name = sub if name == "" else name + "." + sub
                    missed_name = True
                    for d in self:
                        if (
                            p is not d
                            and p.efuse_block == d.efuse_block
                            and name == d.field_name
                        ):
                            missed_name = False
                    if missed_name:
                        raise InputError(f"{name} is not found")

    def verify(self, type_table=None):
        def check(p, n):
            left = n.bit_start
            right = n.bit_start + n.bit_count - 1
            start = p.bit_start
            end = p.bit_start + p.bit_count - 1
            if left <= start <= right:
                if left <= end <= right:
                    return "included in"  # [n  [p...p]  n]
                return "intersected with"  # [n  [p..n]..p]
            if left <= end <= right:
                return "intersected with"  # [p..[n..p] n]
            if start <= left and right <= end:
                return "wraps"  # [p  [n...n]  p]
            return "ok"  # [p] [n]  or  [n] [p]

        def print_error(p, n, state):
            raise InputError(
                f"Field at {p.field_name}, {p.efuse_block}, {p.bit_start}, "
                f"{p.bit_count} {state} {n.field_name}, {n.efuse_block}, "
                f"{n.bit_start}, {n.bit_count}"
            )

        for p in self:
            p.verify(type_table)

        self.verify_duplicate_name()
        if type_table != "custom_table":
            # check will be done for common and custom tables together
            self.check_struct_field_name()

        # check for overlaps
        for p in self:
            for n in self:
                if p is not n and p.efuse_block == n.efuse_block:
                    state = check(p, n)
                    if state != "ok":
                        if "." in p.field_name:
                            name = ""
                            for sub in p.field_name.split("."):
                                name = sub if name == "" else name + "." + sub
                                for d in self:
                                    if (
                                        p is not d
                                        and p.efuse_block == d.efuse_block
                                        and name == d.field_name
                                    ):
                                        state = check(p, d)
                                        if state == "included in":
                                            break
                                        elif state != "intersected with":
                                            state = "out of range"
                                        print_error(p, d, state)
                            continue
                        elif "." in n.field_name:
                            continue
                        print_error(p, n, state)


class FuseDefinition:
    def __init__(self):
        self.field_name = ""
        self.group = ""
        self.efuse_block = ""
        self.bit_start = None
        self.bit_count = None
        self.define = None
        self.comment = ""
        self.alt_names = ""
        self.MAX_BITS_OF_BLOCK = 256

    @classmethod
    def from_csv(cls, line):
        """Parse a line from the CSV"""
        line_w_defaults = line + ",,,,"
        fields = [f.strip() for f in line_w_defaults.split(",")]

        res = FuseDefinition()
        res.field_name = fields[0]
        res.efuse_block = res.parse_block(fields[1])
        res.bit_start = res.parse_num(fields[2])
        res.bit_count = res.parse_bit_count(fields[3])
        if res.bit_count is None or res.bit_count == 0:
            raise InputError("Field bit_count can't be empty")
        res.comment = fields[4].rstrip("\\").rstrip()
        res.comment += f" ({res.bit_start}-{res.bit_start + res.bit_count - 1})"
        res.alt_names = res.get_alt_names(res.comment)
        return res

    def parse_num(self, strval):
        if strval == "":
            return None
        return self.parse_int(strval)

    def parse_bit_count(self, strval):
        if strval == "MAX_BLK_LEN":
            self.define = strval
            return self.MAX_BITS_OF_BLOCK
        else:
            return self.parse_num(strval)

    def parse_int(self, v):
        try:
            return int(v, 0)
        except ValueError:
            raise InputError(f"Invalid field value {v}")

    def parse_block(self, strval):
        if strval == "":
            raise InputError("Field 'efuse_block' can't be left empty.")
        return self.parse_int(strval.lstrip("EFUSE_BLK"))

    def verify(self, type_table):
        if self.efuse_block is None:
            raise ValidationError(self, "efuse_block field is not set")
        if self.bit_count is None:
            raise ValidationError(self, "bit_count field is not set")
        max_bits = self.MAX_BITS_OF_BLOCK
        if self.bit_start + self.bit_count > max_bits:
            raise ValidationError(
                self,
                f"The field is outside the boundaries (max_bits = {max_bits}) "
                f"of the {self.efuse_block} block",
            )

    def get_bit_count(self, check_define=True):
        if check_define is True and self.define is not None:
            return self.define
        else:
            return self.bit_count

    def get_alt_names(self, comment):
        result = re.search(r"^\[(.*?)\]", comment)
        if result:
            return result.group(1)
        return ""


class InputError(RuntimeError):
    def __init__(self, e):
        super().__init__(e)


class ValidationError(InputError):
    def __init__(self, p, message):
        super().__init__(f"Entry {p.field_name} invalid: {message}")



# This file includes the common operations with eFuses for chips
#
# SPDX-FileCopyrightText: 2020-2025 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: GPL-2.0-or-later

import io
import json
import os
import sys
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, BinaryIO, TextIO, cast

import rich_click as click
from bitstring import BitStream

import espsecure
import esptool
from esptool.logger import log

from . import base_fields, util
from .emulate_efuse_controller_base import EmulateEfuseControllerBase


class EfuseArgument(click.Argument):
    def make_metavar(self, ctx: click.Context | None = None) -> str:
        """Compatibility layer for Click 8.2.0+; which now requires a ctx parameter."""
        try:
            return super().make_metavar(ctx)  # type: ignore
        except TypeError:
            # Fall back to the old signature (pre-Click 8.2.0)
            return super().make_metavar()  # type: ignore


class EfuseValuePairArg(EfuseArgument):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def make_metavar(self, ctx=None) -> str:
        return f"[{super().make_metavar(ctx)}] ..."

    def type_cast_value(self, ctx: click.Context, value: list[str]):
        return self.type.convert(value, None, ctx)


class EfuseValuePairType(click.ParamType):
    name = "efuse-value-pair"

    def __init__(self, efuse_choices, efuses):
        self.efuse_choices = efuse_choices
        self.efuses = efuses

    def convert(self, value: str, param: click.Parameter | None, ctx: click.Context):
        def check_efuse_name(efuse_name: str):
            if efuse_name not in self.efuse_choices:
                raise click.BadParameter(
                    f"Invalid eFuse name '{efuse_name}'. "
                    f"Available eFuse names: {self.efuse_choices}"
                )
            return efuse_name

        # Handle single value case (eFuse name only)
        efuse_value_pairs = {}
        if len(value) == 0:
            raise click.BadParameter("Missing eFuse name and value pair.")
        elif len(value) > 1:
            if len(value) % 2:
                raise click.BadParameter(
                    f"The list does not have a valid pair (name value) {value}"
                )
            for i in range(0, len(value), 2):
                efuse_name: str = value[i]
                new_value: str = value[i + 1]
                efuse_name = check_efuse_name(efuse_name)
                check_arg = base_fields.CheckArgValue(self.efuses, efuse_name)
                efuse_value_pairs[efuse_name] = check_arg(new_value)

        else:
            # For the case of compatibility, when only the efuse_name is given
            # Fields with 'bitcount' and 'bool' types can be without new_value arg
            efuse_name = value[0]
            check_efuse_name(efuse_name)
            check_arg = base_fields.CheckArgValue(self.efuses, efuse_name)
            efuse_value_pairs[efuse_name] = check_arg(None)

        return efuse_value_pairs


class CustomMACType(click.ParamType):
    name = "custom_mac"

    def convert(self, value: str, param: click.Parameter | None, ctx: click.Context):
        return base_fields.CheckArgValue(ctx.obj["efuses"], "CUSTOM_MAC")(value)


class TupleParameter(EfuseArgument):
    def __init__(self, *args, **kwargs):
        self.max_arity = kwargs.pop("max_arity", None)
        super().__init__(*args, **kwargs)

    def make_metavar(self, ctx=None) -> str:
        if self.nargs == 1:
            return super().make_metavar(ctx)  # type: ignore
        if self.max_arity is None:
            return f"[{super().make_metavar(ctx)}] ..."
        return f"[{super().make_metavar(ctx)}] ... (max {self.max_arity} groups)"

    def type_cast_value(self, ctx: click.Context, value: list[str]) -> tuple[Any, ...]:
        # This is by default eating all options, so we need to check for help option
        if any(v in ctx.help_option_names for v in value):
            # show help
            click.echo(ctx.get_help())
            ctx.exit()

        # Check if we have more values than allowed by max_arity
        if self.max_arity is not None and len(value) > self.max_arity * self.type.arity:
            raise click.BadParameter(
                f"Expected at most {self.max_arity} groups ({self.type.arity} values "
                f"each), got {len(value)} (values: {value})"
            )

        # check that the number of values is a multiple of self.type.arity
        if len(value) % self.type.arity != 0:
            raise click.BadParameter(
                f"Expected multiple of {self.type.arity} values, got {len(value)} "
                f"(values: {value})"
            )

        # split value into groups of self.type.arity and call convert() for each group
        groups = [
            value[i : i + self.type.arity]
            for i in range(0, len(value), self.type.arity)
        ]
        return tuple(self.type.convert(group, None, ctx) for group in groups)


class NonCompositeTuple(click.Tuple):
    is_composite = False  # Hack to work around click's default nargs=1

    def __init__(self, types):
        super().__init__(types)


def add_force_write_always(function: Callable):
    def callback(ctx: click.Context, param: click.Parameter, value: str):
        ctx.ensure_object(dict)
        if ctx.obj.get("commands", None) is not None:
            ctx.obj["commands"].efuses.force_write_always = value

    return click.option(
        "--force-write-always",
        help="Write the eFuse even if it looks like it's already been written, "
        "or is write protected. Note that this option can't disable write protection, "
        "or clear any bit which has already been set.",
        is_flag=True,
        callback=callback,
    )(function)


def add_show_sensitive_info_option(function: Callable):
    def callback(ctx: click.Context, param: click.Parameter, value: bool):
        if value or ctx.obj.get("debug") or ctx.obj.get("show_sensitive_info", False):
            value = True
        ctx.obj["show_sensitive_info"] = value
        return value

    return click.option(
        "--show-sensitive-info",
        help="Show data to be burned (may expose sensitive data). "
        "Enabled if --debug is used.",
        is_flag=True,
        callback=callback,
        expose_value=True,  # ensure that callback is called even if option is not used
    )(function)


def protect_options(function: Callable):
    function = click.option(
        "--no-write-protect",
        help="Disable write-protecting of the key. The key remains writable. "
        "(The keys use the RS coding scheme that does not support "
        "post-write data changes. Forced write can damage RS encoding bits.) "
        "The write-protecting of keypurposes does not depend on the option, "
        "it will be set anyway.",
        is_flag=True,
    )(function)
    function = click.option(
        "--no-read-protect",
        help="Disable read-protecting of the key. The key remains readable software.",
        is_flag=True,
    )(function)
    return function


class BaseCommands(ABC):
    CHIP_NAME = "auto"
    efuse_lib: type[base_fields.EspEfusesBase] | None = None
    efuses: base_fields.EspEfusesBase
    esp: esptool.ESPLoader | EmulateEfuseControllerBase
    external_esp: bool = False

    def get_efuses(
        self,
        skip_connect=False,
        debug_mode=False,
        do_not_confirm=False,
        extend_efuse_table=None,
    ):
        if self.esp is None:
            raise esptool.FatalError("get_efuses: esp is not set")
        if self.CHIP_NAME == self.esp.CHIP_NAME:
            if self.efuse_lib is None:
                raise esptool.FatalError("get_efuses: efuse_lib is not set")
            self.efuses = self.efuse_lib(
                self.esp, skip_connect, debug_mode, do_not_confirm, extend_efuse_table
            )
        else:
            raise esptool.FatalError(
                "get_efuses: Mismatch chip name "
                f"({self.CHIP_NAME} != {self.esp.CHIP_NAME})"
            )

    @classmethod
    def _close_port(
        cls,
        esp: esptool.ESPLoader | EmulateEfuseControllerBase | None,
        external_esp: bool = False,
    ) -> None:
        """Close the serial port if it was opened internally."""
        if esp is not None and not external_esp and isinstance(esp, esptool.ESPLoader):
            esp._port.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._close_port(self.esp, self.external_esp)

    ################################# CLI definitions #################################

    def add_cli_commands(self, cli: click.Group):
        """Add the CLI commands to the given click group"""

        if self.efuses is None:
            # This should never happen, but just in case someone calls it from API
            raise esptool.FatalError(
                "To initialize the CLI commands, you need to call get_efuses() first."
            )

        efuses: list[base_fields.EfuseFieldBase] = getattr(self.efuses, "efuses", [])

        @cli.command(
            "burn-efuse",
            help="Burn the eFuse with the specified name.\n\n"
            f"Allowed options for EFUSE_NAME: [{', '.join([e.name for e in efuses])}].",
        )
        @click.argument(
            "name_value_pairs",
            cls=EfuseValuePairArg,
            metavar="<EFUSE_NAME> <VALUE>",
            required=True,
            nargs=-1,
            type=EfuseValuePairType(
                [name for e in efuses for name in e.all_names],
                self.efuses,
            ),
        )
        @click.option(
            "--force", is_flag=True, help="Suppress errors when burning eFuses."
        )
        @click.pass_context
        def burn_efuse_cli(ctx, name_value_pairs, force):
            self.burn_efuse(name_value_pairs, force)

        @cli.command(
            "read-protect-efuse",
            help="Disable readback for the selected eFuse with the specified name.",
            short_help="Disable readback for the eFuse.",
        )
        @click.argument(
            "efuse_name",
            nargs=-1,
            required=True,
            type=click.Choice(
                [
                    name
                    for e in efuses
                    if e.read_disable_bit is not None
                    for name in e.all_names
                ]
            ),
        )
        @click.pass_context
        def read_protect_efuse_cli(ctx, efuse_name):
            self.read_protect_efuse(efuse_name)

        @cli.command(
            "write-protect-efuse",
            help="Disable writing to the eFuse with the specified name.",
            short_help="Disable writing to the eFuse.",
        )
        @click.argument(
            "efuse_name",
            nargs=-1,
            required=True,
            type=click.Choice(
                [
                    name
                    for e in efuses
                    if e.write_disable_bit is not None
                    for name in e.all_names
                ]
            ),
        )
        def write_protect_efuse_cli(efuse_name):
            """Disable writing to the eFuse with the specified name."""
            self.write_protect_efuse(efuse_name)

        @cli.command(
            "burn-block-data",
            help="Burn non-key data to EFUSE blocks. "
            "(Don't use this command to burn key data for Flash Encryption or ESP32 "
            "Secure Boot V1, as the byte order of keys is swapped (use burn-key)).\n\n"
            "Allowed options for BLOCK: "
            f"[{', '.join(self.efuses.BURN_BLOCK_DATA_NAMES)}].",
        )
        @click.argument(
            "block_datafile",
            cls=TupleParameter,
            metavar="<BLOCK> <DATAFILE>",
            required=True,
            nargs=-1,
            max_arity=len(self.efuses.BURN_BLOCK_DATA_NAMES),
            type=NonCompositeTuple(
                [
                    click.Choice(self.efuses.BURN_BLOCK_DATA_NAMES),
                    click.File("rb"),
                ]
            ),
        )
        @click.option(
            "--offset",
            "-o",
            type=int,
            default=0,
            help="Byte offset in the eFuse block.",
        )
        @add_force_write_always
        def burn_block_data_cli(block_datafile, offset, **kwargs):
            block, datafile = zip(*block_datafile)
            self.burn_block_data(block, datafile, offset)

        @cli.command("burn-bit")
        @click.argument(
            "block",
            required=True,
            type=click.Choice(self.efuses.BURN_BLOCK_DATA_NAMES),
        )
        @click.argument(
            "bit_number",
            nargs=-1,
            type=int,
            required=True,
        )
        @add_force_write_always
        def burn_bit_cli(block, bit_number, **kwargs):
            """Burn bit in the eFuse block."""
            self.burn_bit(block, bit_number)

        @cli.command("dump")
        @click.option(
            "--format",
            type=click.Choice(["default", "split", "joint"]),
            default="default",
            help="Select the dump format: default - usual console eFuse dump; "
            "joint - all eFuse blocks are stored in one file; "
            "split - each eFuse block is placed into its own file.",
        )
        @click.option(
            "--file-name",
            type=click.Path(dir_okay=False, writable=True),
            default=None,
            help="The path to the file in which to save the dump, if not specified, "
            "output to the console.",
        )
        def dump_cli(format, file_name):
            """Dump raw hex values of all eFuses."""
            self.dump(format, file_name)

        @cli.command("summary")
        @click.argument("efuses_to_show", nargs=-1, required=False)
        @click.option(
            "--format",
            type=click.Choice(["summary", "json", "value_only"]),
            default="summary",
            help="Select the summary format.",
        )
        @click.option(
            "--file",
            type=click.File("w"),
            default=sys.stdout,
            help="File to save the eFuse summary to.",
        )
        def summary_cli(format, file, efuses_to_show=[]):
            """Print human-readable summary of eFuse values."""
            self.summary(efuses_to_show, format, file)

        @cli.command("check-error")
        @click.option(
            "--recovery", is_flag=True, help="Recovery of BLOCKs after encoding errors."
        )
        @click.pass_context
        def check_error_cli(ctx, recovery):
            """Checks eFuse errors."""
            self.check_error(recovery, ctx.obj["do_not_confirm"])

        @cli.command(
            "adc-info",
            short_help="Display information about ADC calibration data "
            "stored in eFuse.",
            help="Display information about ADC calibration data stored in eFuse.",
        )
        def adc_info_cli():
            self.adc_info()

        @cli.command(
            "burn-custom-mac",
            short_help="Burn a 48-bit Custom MAC address.",
            help="Burn a 48-bit Custom MAC address to EFUSE,"
            f"BLOCK{self.efuses['CUSTOM_MAC'].block}. "
            "Mac address should be given in hexadecimal format with bytes separated "
            "by colons (e.g. AA:CD:EF:01:02:03).",
        )
        @click.argument(
            "mac",
            type=CustomMACType(),
        )
        @add_force_write_always
        def burn_custom_mac_cli(mac, **kwargs):
            self.burn_custom_mac(mac)

        @cli.command("get-custom-mac")
        def get_custom_mac_cli():
            """Get the 48-bit Custom MAC Address."""
            self.get_custom_mac()

    ################################## Helper methods ##################################

    def use_batch_mode(self):
        """Enable batch mode for eFuse operations.

        This method increments the batch mode counter, allowing multiple eFuse
        operations to be queued before burning. All queued operations will be
        executed together when :func:`burn_all` is called.

        This method can be called multiple times to nest batch operations.
        Each call should be paired with a corresponding call to :func:`burn_all`.
        """
        self.efuses.batch_mode_cnt += 1

    def burn_all(self, check_batch_mode=False):
        """Execute all queued eFuse operations and decrement batch mode counter.

        This method decrements the batch mode counter, then burns all eFuses
        that have been queued during batch mode. When the counter reaches zero,
        batch mode is fully exited.

        This method should be called after :func:`use_batch_mode` to execute
        the queued operations.

        Args:
            check_batch_mode: If True, only execute the burn operation if this is the
                final burn_all call (i.e., when batch_mode_cnt reaches zero in nested
                operations).

        Returns:
            bool: True if the burn operation was successful.
        """
        self.efuses.batch_mode_cnt -= 1
        return self.efuses.burn_all(check_batch_mode)

    def _key_block_is_unused(
        self,
        block: base_fields.EfuseBlockBase,
        key_purpose_block: base_fields.EfuseFieldBase,
    ) -> bool:
        """Helper method to check if a key block is available for use"""
        if not block.is_readable() or not block.is_writeable():
            return False

        if key_purpose_block.get() != "USER" or not key_purpose_block.is_writeable():
            return False

        if not block.get_bitstring().all(False):
            return False

        return True

    def _get_next_key_block(
        self, current_key_block: base_fields.EfuseBlockBase, block_name_list: list[str]
    ) -> base_fields.EfuseBlockBase | None:
        """Helper method to get the next available key block"""
        key_blocks = [b for b in self.efuses.blocks if b.key_purpose_name]
        start = key_blocks.index(current_key_block)

        # Sort key blocks so that we pick the next free block
        # (and loop around if necessary)
        key_blocks = key_blocks[start:] + key_blocks[0:start]

        # Exclude any other blocks that will be be burned
        key_blocks = [b for b in key_blocks if b.name not in block_name_list]

        for block in key_blocks:
            key_purpose_block = self.efuses[cast(str, block.key_purpose_name)]
            if self._key_block_is_unused(block, key_purpose_block):
                return block

        return None

    def _adjust_key_data_for_blocks(
        self,
        block_names: list[str],
        datafiles: list[BinaryIO],
        keypurposes: list[str],
    ) -> tuple[list[str], list[BinaryIO], list[str]]:
        """Split a key that takes more than one efuse block into two blocks.
        It handles all key purposes that require splitting into two blocks.

        This method checks if key purposes require splitting into two blocks,
        such as "XTS_AES_256_KEY", "XTS_AES_256_PSRAM_KEY", and "ECDSA_KEY_P384".

        Args:
            block_names: List of block names.
            datafiles: List of BinaryIO objects containing key data.
            keypurposes: List of key purposes.

        Returns:
            A tuple containing updated block names, datafiles, and keypurposes.
        """
        keypurposes = list(keypurposes)
        datafiles = list(datafiles)
        block_names = list(block_names)

        if "XTS_AES_256_KEY" in keypurposes:
            # XTS_AES_256_KEY is not an actual HW key purpose, needs to be split into
            # XTS_AES_256_KEY_1 and XTS_AES_256_KEY_2
            block_names, datafiles, keypurposes = self._split_multiblock_key(
                block_names,
                datafiles,  # type: ignore
                keypurposes,
                "XTS_AES_256_KEY",
            )

        if "XTS_AES_256_PSRAM_KEY" in keypurposes:
            # XTS_AES_256_PSRAM_KEY -> XTS_AES_256_PSRAM_KEY_1 and ..._KEY_2
            block_names, datafiles, keypurposes = self._split_multiblock_key(
                block_names,
                datafiles,  # type: ignore
                keypurposes,
                "XTS_AES_256_PSRAM_KEY",
            )

        # ECDSA keys can be present in a command multiple times
        i = 0
        while i < len(keypurposes):
            if "ECDSA_KEY" in keypurposes[i]:
                if keypurposes[i] not in ["ECDSA_KEY_P384_L", "ECDSA_KEY_P384_H"]:
                    sk = espsecure.load_ecdsa_signing_key(datafiles[i])  # type: ignore
                    data = espsecure.get_ecdsa_signing_key_raw_bytes(sk)
                    if "ECDSA_KEY_P384" == keypurposes[i]:
                        assert len(data) == 48, (
                            "NIST384p private key should be 48 bytes long"
                        )
                        datafiles[i] = io.BytesIO(b"\x00" * 16 + data)
                        # ECDSA_KEY_P384 -> ECDSA_KEY_P384_L and ECDSA_KEY_P384_H
                        block_names, datafiles, keypurposes = (
                            self._split_multiblock_key(
                                block_names,
                                datafiles,  # type: ignore
                                keypurposes,
                                "ECDSA_KEY_P384",
                            )
                        )
                    else:
                        # the private key is 24 bytes long for NIST192p,
                        # and 8 bytes of padding
                        datafiles[i] = (
                            io.BytesIO(b"\x00" * 8 + data)
                            if len(data) == 24
                            else io.BytesIO(data)
                        )

            i += 1

        # Check that all block names are unique
        util.check_duplicate_name_in_list(block_names)

        # Check that the number of blocks, datafiles, and keypurposes is equal
        if len(block_names) != len(datafiles) or len(block_names) != len(keypurposes):
            raise esptool.FatalError(
                f"The number of blocks ({len(block_names)}), "
                f"datafile ({len(datafiles)}) and keypurpose ({len(keypurposes)}) "
                "should be the same."
            )

        return block_names, datafiles, keypurposes

    def _split_multiblock_key(
        self,
        block_names: list[str],
        datafiles: list[BinaryIO],
        keypurposes: list[str],
        base_keypurpose: str,
    ) -> tuple[list[str], list[BinaryIO], list[str]]:
        """Helper method to split 512-bit key into two 256-bit keys"""
        keypurpose_list = list(keypurposes)
        datafile_list = list(datafiles)
        block_name_list = list(block_names)

        i = keypurpose_list.index(base_keypurpose)
        block_name = block_name_list[i]

        block_num = self.efuses.get_index_block_by_name(block_name)
        block = self.efuses.blocks[block_num]

        data = datafile_list[i].read()
        if len(data) != 64:
            raise esptool.FatalError(
                f"Incorrect key file size {len(data)}, {base_keypurpose} "
                "should be 64 bytes"
            )

        key_block_2 = self._get_next_key_block(block, block_name_list)
        if not key_block_2:
            raise esptool.FatalError(f"{base_keypurpose} requires two free keyblocks")

        postfix = (
            ["_1", "_2"] if base_keypurpose.startswith("XTS_AES_256") else ["_H", "_L"]
        )
        keypurpose_list[i] = f"{base_keypurpose}{postfix[0]}"
        datafile_list[i] = io.BytesIO(data[:32])
        block_name_list[i] = block_name

        keypurpose_list.insert(i + 1, f"{base_keypurpose}{postfix[1]}")
        datafile_list.insert(i + 1, io.BytesIO(data[32:]))
        block_name_list.insert(i + 1, key_block_2.name)

        return block_name_list, datafile_list, keypurpose_list

    def _convert_mac_to_bytes(self, mac: str | bytes) -> bytes:
        if isinstance(mac, str):
            return base_fields.CheckArgValue(self.efuses, "CUSTOM_MAC")(mac)  # type: ignore
        return mac

    ##################################### Commands ####################################

    def summary(
        self,
        efuses_to_show: list[str] = [],
        format: str = "summary",
        file: TextIO = sys.stdout,
    ):
        """
        Print a human-readable or json summary of eFuse contents.

        Args:
            efuses_to_show: List of eFuse names to show.
            format: Format to use for the summary.
            file: File to write the summary to.
        """
        ROW_FORMAT = "%-50s %-50s%s = %s %s %s"
        human_output = format in ["summary", "value_only"]
        value_only = format == "value_only"
        if value_only and len(efuses_to_show) != 1:
            raise esptool.FatalError(
                "The 'value_only' format can be used exactly for one eFuse."
            )
        do_filtering = bool(efuses_to_show)
        json_efuse = {}
        summary_efuse = []
        if file != sys.stdout:
            log.print("Saving eFuse values to " + file.name)
        if human_output and not value_only:
            summary_efuse.append(
                ROW_FORMAT.replace("-50", "-12")
                % (
                    "EFUSE_NAME (Block)",
                    "Description",
                    "",
                    "[Meaningful Value]",
                    "[Readable/Writeable]",
                    "(Hex Value)",
                )
            )
            summary_efuse.append("-" * 88)
        for category in sorted(
            set(e.category for e in self.efuses), key=lambda c: c.title()
        ):
            if human_output and not value_only:
                summary_efuse.append(f"{category.title()} fuses:")
            for e in (e for e in self.efuses if e.category == category):
                if e.efuse_type.startswith("bytes"):
                    raw = ""
                else:
                    raw = "({})".format(e.get_bitstring())
                (readable, writeable) = (e.is_readable(), e.is_writeable())
                if readable and writeable:
                    perms = "R/W"
                elif readable:
                    perms = "R/-"
                elif writeable:
                    perms = "-/W"
                else:
                    perms = "-/-"
                base_value = e.get_meaning()
                value = str(base_value)
                if not readable:
                    count_read_disable_bits = e.get_count_read_disable_bits()
                    if count_read_disable_bits == 2:
                        # On the C2 chip, BLOCK_KEY0 has two read protection bits [0, 1]
                        # related to the lower and higher part of the block.
                        v = [value[: (len(value) // 2)], value[(len(value) // 2) :]]
                        for i in range(count_read_disable_bits):
                            if not e.is_readable(blk_part=i):
                                v[i] = v[i].replace("0", "?")
                        value = "".join(v)
                    else:
                        value = value.replace("0", "?")
                if (
                    human_output
                    and (not do_filtering or e.name in efuses_to_show)
                    and not value_only
                ):
                    summary_efuse.append(
                        ROW_FORMAT
                        % (
                            e.get_info(),
                            e.description[:50],
                            "\n  " if len(value) > 20 else "",
                            value,
                            perms,
                            raw,
                        )
                    )
                    desc_len = len(e.description[50:])
                    if desc_len:
                        desc_len += 50
                        for i in range(50, desc_len, 50):
                            summary_efuse.append(
                                f"{'':<50} {e.description[i : (50 + i)]:<50}"
                            )
                elif human_output and value_only and e.name in efuses_to_show:
                    summary_efuse.append(f"{value}")
                elif format == "json" and (
                    not do_filtering or e.name in efuses_to_show
                ):
                    json_efuse[e.name] = {
                        "name": e.name,
                        "value": base_value if readable else value,
                        "raw_value": util.json_raw_value_hex(
                            e.efuse_type, e.get_bitstring()
                        ),
                        "readable": readable,
                        "writeable": writeable,
                        "description": e.description,
                        "category": e.category,
                        "block": e.block,
                        "word": e.word,
                        "pos": e.pos,
                        "efuse_type": e.efuse_type,
                        "bit_len": e.bit_len,
                    }
            if human_output and not value_only:
                # Remove empty category when filtered efuses have none to show
                if do_filtering and summary_efuse[-1] == f"{category.title()} fuses:":
                    summary_efuse.pop()
                else:
                    summary_efuse.append("")
        if human_output and not value_only:
            summary_efuse.append(self.efuses.summary())
            warnings = self.efuses.get_coding_scheme_warnings()
            if warnings:
                summary_efuse.append(
                    "WARNING: Coding scheme has encoding bit error warnings"
                )
        if human_output:
            for line in summary_efuse:
                log.print(line, file=file)
            if file != sys.stdout:
                file.close()
                log.print("Done")
        elif format == "json":
            json.dump(json_efuse, file, sort_keys=True, indent=4)
            log.print("")

    def dump(self, format: str = "default", file_name: str | None = None):
        """
        Dump raw eFuse data registers.

        Args:
            format: Format to use for the dump. Available options are:
                - "default": Print the dump to the console.
                - "split": Dump each eFuse block to a separate file.
                - "joint": Dump all eFuse blocks to a single file.
            file_name: File to write the dump to. If not provided, the dump will
                be printed to the console.
        """
        to_console = file_name is None
        dump_file: BinaryIO | TextIO = sys.stdout

        def output_block_to_file(
            block: base_fields.EfuseBlockBase, f: BinaryIO | TextIO, to_console: bool
        ):
            block_dump = BitStream(block.get_bitstring())
            block_dump.byteswap()
            if to_console:
                f.write(block_dump.hex + "\n")
            else:
                block_dump.tofile(f)  # type: ignore

        if format == "default":
            if to_console:
                # for "espefuse dump" cmd
                for block in self.efuses.blocks:
                    block.print_block(block.get_bitstring(), "dump", debug=True)
                return
            else:
                # for back compatibility to support
                # "espefuse dump --file_name dump.bin"
                format = "split"

        if format == "split":
            # each eFuse block is placed into its own file
            for block in self.efuses.blocks:
                if not to_console:
                    fname, fextension = os.path.splitext(file_name)  # type: ignore
                    file_dump_name = f"{fname}{block.id}{fextension}"
                    log.print(f"Dump eFuse block{block.id} -> {file_dump_name}")
                    dump_file = open(file_dump_name, "wb")
                output_block_to_file(block, dump_file, to_console)
                if not to_console:
                    dump_file.close()
        elif format == "joint":
            # all eFuse blocks are stored in one file
            if not to_console:
                log.print(f"Dump eFuse blocks -> {file_name}")
                dump_file = open(file_name, "wb")  # type: ignore
            for block in self.efuses.blocks:
                output_block_to_file(block, dump_file, to_console)
            if not to_console:
                dump_file.close()

    def burn_efuse(self, name_value_pairs: dict[str, str], force: bool = False):
        """
        Burn eFuses.

        Args:
            name_value_pairs: Dictionary of eFuse names and values to burn.
            force: If True, the burn will be performed even if the eFuse settings
                are incompatible.
        """

        def print_attention(blocked_efuses_after_burn: list[str]):
            if len(blocked_efuses_after_burn):
                log.print(
                    "    ATTENTION! This BLOCK uses NOT the NONE coding scheme "
                    "and after 'BURN', these efuses can not be burned in the feature:"
                )
                for i in range(0, len(blocked_efuses_after_burn), 5):
                    log.print(
                        "              ",
                        "".join("{}".format(blocked_efuses_after_burn[i : i + 5 :])),
                    )

        efuse_name_list = [name for name in name_value_pairs.keys()]
        burn_efuses_list = [self.efuses[name] for name in efuse_name_list]
        old_value_list = [self.efuses[name].get_raw() for name in efuse_name_list]
        new_value_list = [value for value in name_value_pairs.values()]
        util.check_duplicate_name_in_list(efuse_name_list)

        attention = ""
        log.print("The eFuses to burn:")
        for block in self.efuses.blocks:
            burn_list_a_block = [e for e in burn_efuses_list if e.block == block.id]
            if len(burn_list_a_block):
                log.print(f"  from BLOCK{block.id}")
                for field in burn_list_a_block:
                    log.print(f"     - {field.name}")
                    if (
                        self.efuses.blocks[field.block].get_coding_scheme()
                        != self.efuses.REGS.CODING_SCHEME_NONE
                    ):
                        using_the_same_block_names = [
                            e.name for e in self.efuses if e.block == field.block
                        ]
                        wr_names = [e.name for e in burn_list_a_block]
                        blocked_efuses_after_burn = [
                            name
                            for name in using_the_same_block_names
                            if name not in wr_names
                        ]
                        attention = " (see 'ATTENTION!' above)"
                if attention:
                    print_attention(blocked_efuses_after_burn)

        log.print(f"\nBurning eFuses{attention}:")
        for efuse, new_value in zip(burn_efuses_list, new_value_list):
            log.print(
                f"    - '{efuse.name}' ({efuse.description}) "
                f"{efuse.get_bitstring()} -> {efuse.convert_to_bitstring(new_value)}"
            )
            efuse.save(new_value)

        log.print()
        if "ENABLE_SECURITY_DOWNLOAD" in efuse_name_list:
            log.print(
                "ENABLE_SECURITY_DOWNLOAD -> 1: eFuses will not be read back "
                "for confirmation because this mode disables "
                "any SRAM and register operations."
            )
            log.print("                               espefuse will not work.")
            log.print(
                "                               esptool can read/write only flash."
            )

        if "DIS_DOWNLOAD_MODE" in efuse_name_list:
            log.print(
                "DIS_DOWNLOAD_MODE -> 1: eFuses will not be read back for confirmation "
                "because this mode disables any communication with the chip."
            )
            log.print(
                "                        espefuse/esptool will not work because "
                "they will not be able to connect to the chip."
            )

        if (
            self.esp.CHIP_NAME == "ESP32"
            and self.esp.get_chip_revision() >= 300
            and "UART_DOWNLOAD_DIS" in efuse_name_list
        ):
            log.print(
                "UART_DOWNLOAD_DIS -> 1: eFuses will be read for confirmation, "
                "but after that connection to the chip will become impossible."
            )
            log.print("                        espefuse/esptool will not work.")

        if self.efuses.is_efuses_incompatible_for_burn():
            if force:
                log.print("Ignore incompatible eFuse settings.")
            else:
                raise esptool.FatalError(
                    "Incompatible eFuse settings detected, abort. "
                    "(use --force flag to skip it)."
                )

        if not self.efuses.burn_all(check_batch_mode=True):
            return

        log.print("Checking eFuses...")
        raise_error = False
        for efuse, old_value, new_value in zip(
            burn_efuses_list, old_value_list, new_value_list
        ):
            if not efuse.is_readable():
                log.print(
                    f"Efuse {efuse.name} is read-protected. "
                    "Read back the burn value is not possible."
                )
            else:
                new_value = efuse.convert_to_bitstring(new_value)
                burned_value = efuse.get_bitstring()
                if burned_value != new_value:
                    log.print(
                        burned_value,
                        "->",
                        new_value,
                        f"Efuse {efuse.name} failed to burn. Protected?",
                    )
                    raise_error = True
        if raise_error:
            raise esptool.FatalError("The burn was not successful.")
        else:
            log.print("Successful.")

    def read_protect_efuse(self, efuse_names: list[str]):
        """
        Disable readback for the eFuse with the specified name.

        Args:
            efuse_names: List of eFuse names to read-protect.
        """
        util.check_duplicate_name_in_list(efuse_names)

        for efuse_name in efuse_names:
            efuse = self.efuses[efuse_name]
            if not efuse.is_readable():
                log.print(f"Efuse {efuse.name} is already read protected")
            else:
                if self.esp.CHIP_NAME == "ESP32":
                    if (
                        efuse_name == "BLOCK2"
                        and not self.efuses["ABS_DONE_0"].get()
                        and self.esp.get_chip_revision() >= 300
                    ):
                        if self.efuses["ABS_DONE_1"].get():
                            raise esptool.FatalError(
                                "Secure Boot V2 is on (ABS_DONE_1 = True), "
                                "BLOCK2 must be readable, stop this operation!"
                            )
                        else:
                            log.print(
                                "If Secure Boot V2 is used, BLOCK2 must be readable, "
                                "please stop this operation!"
                            )
                elif self.esp.CHIP_NAME == "ESP32-C2":
                    error = (
                        not self.efuses["XTS_KEY_LENGTH_256"].get()
                        and efuse_name == "BLOCK_KEY0"
                    )
                    error |= cast(
                        bool, self.efuses["SECURE_BOOT_EN"].get()
                    ) and efuse_name in ["BLOCK_KEY0", "BLOCK_KEY0_HI_128"]
                    if error:
                        raise esptool.FatalError(
                            f"{efuse_name} must be readable, stop this operation!"
                        )
                else:
                    for block_tuple in self.efuses.Blocks.BLOCKS:
                        block = self.efuses.Blocks.get(block_tuple)
                        if block.name == efuse_name and block.key_purpose is not None:
                            if not self.efuses[block.key_purpose].need_rd_protect(
                                cast(str, self.efuses[block.key_purpose].get())
                            ):
                                raise esptool.FatalError(
                                    f"{efuse_name} must be readable, "
                                    f"stop this operation!"
                                )
                            break
                # make full list of which efuses will be disabled
                # (ie share a read disable bit)
                all_disabling = [
                    e
                    for e in self.efuses
                    if e.read_disable_bit == efuse.read_disable_bit
                ]
                names = ", ".join(e.name for e in all_disabling)
                log.print(
                    f"Permanently read-disabling eFuse"
                    f"{'s' if len(all_disabling) > 1 else ''} {names}"
                )
                efuse.disable_read()

        if not self.efuses.burn_all(check_batch_mode=True):
            return

        log.print("Checking eFuses...")
        raise_error = False
        for efuse_name in efuse_names:
            efuse = self.efuses[efuse_name]
            if efuse.is_readable():
                log.print(f"Efuse {efuse.name} is not read-protected.")
                raise_error = True
        if raise_error:
            raise esptool.FatalError("The burn was not successful.")
        else:
            log.print("Successful.")

    def write_protect_efuse(self, efuse_names: list[str]):
        """
        Disable writing to the eFuse with the specified name.

        Args:
            efuse_names: List of eFuse names to write-protect.
        """
        util.check_duplicate_name_in_list(efuse_names)
        for efuse_name in efuse_names:
            efuse = self.efuses[efuse_name]
            if not efuse.is_writeable():
                log.print(f"Efuse {efuse.name} is already write protected.")
            else:
                # make full list of which efuses will be disabled
                # (ie share a write disable bit)
                all_disabling = [
                    e
                    for e in self.efuses
                    if e.write_disable_bit == efuse.write_disable_bit
                ]
                names = ", ".join(e.name for e in all_disabling)
                log.print(
                    f"Permanently write-disabling eFuse"
                    f"{'s' if len(all_disabling) > 1 else ''} {names}"
                )
                efuse.disable_write()

        if not self.efuses.burn_all(check_batch_mode=True):
            return

        log.print("Checking eFuses...")
        raise_error = False
        for efuse_name in efuse_names:
            efuse = self.efuses[efuse_name]
            if efuse.is_writeable():
                log.print(f"Efuse {efuse.name} is not write-protected.")
                raise_error = True
        if raise_error:
            raise esptool.FatalError("The burn was not successful.")
        else:
            log.print("Successful.")

    def burn_block_data(
        self,
        block_names: list[str],
        datafiles: list[BinaryIO],
        offset: int = 0,
    ) -> None:
        """
        Burn non-key data to EFUSE blocks.

        Don't use this command to burn key data for Flash Encryption or ESP32
        Secure Boot V1, as the byte order of keys is swapped (use burn-key).

        Args:
            block_names: List of eFuse block names to burn data to.
            datafiles: List of files to read data from.
            offset: Byte offset in the eFuse block to start writing data at.
        """
        block_name_list = block_names[
            0 : len([name for name in block_names if name is not None]) :
        ]
        datafile_list = datafiles[
            0 : len([name for name in datafiles if name is not None]) :
        ]

        util.check_duplicate_name_in_list(block_name_list)
        if offset and len(block_name_list) > 1:
            raise esptool.FatalError(
                "The 'offset' option is not applicable when a few blocks are passed. "
                "With 'offset', should only one block be used."
            )
        else:
            if offset:
                num_block = self.efuses.get_index_block_by_name(block_name_list[0])
                block: base_fields.EfuseBlockBase = self.efuses.blocks[num_block]
                num_bytes = block.get_block_len()
                if offset >= num_bytes:
                    raise esptool.FatalError(
                        f"Invalid offset: the block{block.id} only holds "
                        f"{num_bytes} bytes."
                    )
        if len(block_name_list) != len(datafile_list):
            raise esptool.FatalError(
                f"The number of block_name ({len(block_name_list)}) and "
                f"datafile ({len(datafile_list)}) should be the same."
            )

        for block_name, datafile in zip(block_name_list, datafile_list):
            num_block = self.efuses.get_index_block_by_name(block_name)
            block = self.efuses.blocks[num_block]
            data = datafile.read()
            num_bytes = block.get_block_len()
            if offset != 0:
                data = (b"\x00" * offset) + data
                data = data + (b"\x00" * (num_bytes - len(data)))
            if len(data) != num_bytes:
                raise esptool.FatalError(
                    f"Data does not fit: the block{block.id} size is "
                    f"{num_bytes} bytes, data file is {len(data)} bytes, "
                    f"offset {offset}."
                )
            log.print(
                "[{:02}] {:20} size={:02} bytes, offset={:02} - > [{}].".format(
                    block.id, block.name, len(data), offset, util.hexify(data, " ")
                )
            )
            block.save(data)

        if not self.efuses.burn_all(check_batch_mode=True):
            return
        log.print("Successful.")

    def burn_bit(self, block: str, bit_number: list[int]):
        """
        Burn a single bit to the eFuse with the specified name.

        Args:
            block: Name of the eFuse block to burn the bit to.
            bit_number: List of bit numbers to burn.
        """
        num_block = self.efuses.get_index_block_by_name(block)
        block_obj: base_fields.EfuseBlockBase = self.efuses.blocks[num_block]
        data_block = BitStream(block_obj.get_block_len() * 8)
        data_block.set(0)
        try:
            data_block.set(True, bit_number)
        except IndexError:
            raise esptool.FatalError(
                f"{block} has bit_number in [0..{data_block.len - 1}]"
            )
        data_block.reverse()
        log.print(
            f"bit_number:   [{data_block.len - 1:03d}]"
            f"........................................................[0]"
        )
        log.print(f"BLOCK{block_obj.id:>2d}   :", data_block)
        block_obj.print_block(data_block, "regs_to_write", debug=True)
        block_obj.save(data_block.bytes[::-1])

        if not self.efuses.burn_all(check_batch_mode=True):
            return
        log.print("Successful.")

    def get_error_summary(self):
        self.efuses.get_coding_scheme_warnings()
        error_in_blocks = any(
            blk.fail or blk.num_errors != 0 for blk in self.efuses.blocks
        )
        if not error_in_blocks:
            return False
        writable = True
        for blk in self.efuses.blocks:
            if blk.fail or blk.num_errors:
                if blk.id == 0:
                    for field in self.efuses:
                        if field.block == blk.id and (field.fail or field.num_errors):
                            wr = "writable" if field.is_writeable() else "not writable"
                            writable &= wr == "writable"
                            name = field.name
                            val = field.get()
                            log.print(
                                f"BLOCK{field.block:<2}: {name:<40} = {val:<8} ({wr})"
                            )
                else:
                    wr = "writable" if blk.is_writeable() else "not writable"
                    writable &= wr == "writable"
                    name = f"{blk.name} [ERRORS:{blk.num_errors} FAIL:{int(blk.fail)}]"
                    val = str(blk.get_bitstring())
                    log.print(f"BLOCK{blk.id:<2}: {name:<40} = {val:<8} ({wr})")
        if not writable and error_in_blocks:
            log.print(
                "Not all errors can be fixed because some fields are write-protected!"
            )
        return True

    def check_error(self, recovery: bool = False, do_not_confirm: bool = False):
        """
        Check for errors in the eFuse blocks.

        Args:
            recovery: Recovery of BLOCKs after encoding errors.
            do_not_confirm: If True, the confirmation will be skipped.
        """
        error_in_blocks = self.get_error_summary()
        if recovery and error_in_blocks:
            confirmed = False
            for block in reversed(self.efuses.blocks):
                if block.fail or block.num_errors > 0:
                    if not block.get_bitstring().all(False):
                        block.save(block.get_bitstring().bytes[::-1])
                        if not confirmed:
                            confirmed = True
                            self.efuses.confirm(
                                "Recovery of block coding errors", do_not_confirm
                            )
                        block.burn()
            if confirmed:
                self.efuses.update_efuses()
            error_in_blocks = self.get_error_summary()
        if error_in_blocks:
            raise esptool.FatalError("Error(s) were detected in eFuses.")
        log.print("No errors detected.")

    def burn_custom_mac(self, mac: str | bytes):
        """
        Burn a 48-bit Custom MAC Address.

        Args:
            mac (str | bytes): Custom MAC Address to burn. e.g. "aa:cd:ef:11:22:33" or
                b'\xaa\xcd\xef\x11\x22\x33'
        """
        mac = self._convert_mac_to_bytes(mac)

        self.efuses["CUSTOM_MAC"].save(mac)
        if not self.efuses.burn_all(check_batch_mode=True):
            return
        self.get_custom_mac()
        log.print("Successful.")

    def get_custom_mac(self):
        """Get the Custom MAC Address."""
        log.print(f"Custom MAC Address: {self.efuses['CUSTOM_MAC'].get()}")

    def set_flash_voltage(self, voltage: str):
        """
        Set the Flash Voltage. Available only for selected chips.

        Args:
            voltage: Voltage to set. Available options are: "1.8V", "3.3V", "OFF"
        """
        raise esptool.FatalError("set_flash_voltage is not supported for this chip")

    def adc_info(self):
        """Display information about ADC calibration data stored in eFuse."""
        raise NotImplementedError("adc-info is not implemented for this chip")

    @abstractmethod
    def burn_key(self, *args, **kwargs):
        """Burn a key to the eFuse. Exact implementation is chip-specific."""
        pass

    @abstractmethod
    def burn_key_digest(self, *args, **kwargs):
        """Burn a key digest to the eFuse. Exact implementation is chip-specific."""
        pass



# SPDX-FileCopyrightText: 2023-2025 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: GPL-2.0-or-later

from pkcs11.exceptions import (
    AnotherUserAlreadyLoggedIn,
    ArgumentsBad,
    DeviceRemoved,
    DomainParamsInvalid,
    FunctionFailed,
    MechanismInvalid,
    NoSuchKey,
    NoSuchToken,
    OperationNotInitialized,
    SessionClosed,
)

from esptool.logger import log

try:
    # AlreadyInitialized is not available since python-pkcs11 9.0, as multiple
    # initializations are now supported.
    from pkcs11.exceptions import AlreadyInitialized
except ImportError:
    AlreadyInitialized = None


def handle_exceptions(e, info=""):
    exception_type = e.__class__
    if exception_type == MechanismInvalid:
        log.error(f"The External HSM does not support the given mechanism: {info}")
    elif exception_type == FunctionFailed:
        log.error(
            "Please ensure proper configuration, privileges and environment variables."
        )
    elif exception_type == AlreadyInitialized:
        log.error("pkcs11 is already initialized with another library.")
    elif exception_type == AnotherUserAlreadyLoggedIn:
        log.error("Another User has been already logged in.")
    elif exception_type == ArgumentsBad:
        log.error("Please check the arguments supplied to the function.")
    elif exception_type == DomainParamsInvalid:
        log.error(
            "Invalid or unsupported domain parameters were supplied to the function."
        )
    elif exception_type == DeviceRemoved:
        log.error(
            "The token has been removed from its slot during "
            "the execution of the function."
        )
    elif exception_type == NoSuchToken:
        log.error("No such token found.")
    elif exception_type == NoSuchKey:
        log.error("No such key found.")
    elif exception_type == OperationNotInitialized:
        log.error("Operation not initialized.")
    elif exception_type == SessionClosed:
        log.error("Session already closed.")
    else:
        log.error(f"{e.__class__}: {info}")



# SPDX-FileCopyrightText: 2024-2025 Fredrik Ahlberg, Angus Gratton,
# Espressif Systems (Shanghai) CO LTD, other contributors as noted.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from ..loader import ESPLoader, StubMixin
from ..logger import log
from ..util import FatalError
from .esp32c6 import ESP32C6ROM


class ESP32H2ROM(ESP32C6ROM):
    CHIP_NAME = "ESP32-H2"
    IMAGE_CHIP_ID = 16

    DR_REG_LP_WDT_BASE = 0x600B1C00
    RTC_CNTL_WDTCONFIG0_REG = DR_REG_LP_WDT_BASE + 0x0  # LP_WDT_RWDT_CONFIG0_REG
    RTC_CNTL_WDTCONFIG1_REG = DR_REG_LP_WDT_BASE + 0x0004  # LP_WDT_RWDT_CONFIG1_REG
    RTC_CNTL_WDTWPROTECT_REG = DR_REG_LP_WDT_BASE + 0x001C  # LP_WDT_RWDT_WPROTECT_REG

    RTC_CNTL_SWD_CONF_REG = DR_REG_LP_WDT_BASE + 0x0020  # LP_WDT_SWD_CONFIG_REG
    RTC_CNTL_SWD_AUTO_FEED_EN = 1 << 18
    RTC_CNTL_SWD_WPROTECT_REG = DR_REG_LP_WDT_BASE + 0x0024  # LP_WDT_SWD_WPROTECT_REG
    RTC_CNTL_SWD_WKEY = 0x50D83AA1  # LP_WDT_SWD_WKEY, same as WDT key in this case

    FLASH_FREQUENCY = {
        "48m": 0xF,
        "24m": 0x0,
        "16m": 0x1,
        "12m": 0x2,
    }

    UF2_FAMILY_ID = 0x332726F6

    KEY_PURPOSES: dict[int, str] = {
        0: "USER/EMPTY",
        1: "ECDSA_KEY",
        2: "XTS_AES_256_KEY_1",
        3: "XTS_AES_256_KEY_2",
        4: "XTS_AES_128_KEY",
        5: "HMAC_DOWN_ALL",
        6: "HMAC_DOWN_JTAG",
        7: "HMAC_DOWN_DIGITAL_SIGNATURE",
        8: "HMAC_UP",
        9: "SECURE_BOOT_DIGEST0",
        10: "SECURE_BOOT_DIGEST1",
        11: "SECURE_BOOT_DIGEST2",
    }

    def get_pkg_version(self):
        num_word = 4
        return (self.read_reg(self.EFUSE_BLOCK1_ADDR + (4 * num_word)) >> 0) & 0x07

    def get_minor_chip_version(self):
        num_word = 3
        return (self.read_reg(self.EFUSE_BLOCK1_ADDR + (4 * num_word)) >> 18) & 0x07

    def get_major_chip_version(self):
        num_word = 3
        return (self.read_reg(self.EFUSE_BLOCK1_ADDR + (4 * num_word)) >> 21) & 0x03

    def get_chip_description(self):
        chip_name = {
            0: "ESP32-H2",
        }.get(self.get_pkg_version(), "Unknown ESP32-H2")
        major_rev = self.get_major_chip_version()
        minor_rev = self.get_minor_chip_version()
        return f"{chip_name} (revision v{major_rev}.{minor_rev})"

    def get_chip_features(self):
        return ["BT 5 (LE)", "IEEE802.15.4", "Single Core", "96MHz"]

    def get_crystal_freq(self):
        # ESP32H2 XTAL is fixed to 32MHz
        return 32

    # Watchdog reset is not supported on ESP32-H2
    def watchdog_reset(self):
        ESPLoader.watchdog_reset(self)

    def check_spi_connection(self, spi_connection):
        if not set(spi_connection).issubset(set(range(0, 28))):
            raise FatalError("SPI Pin numbers must be in the range 0-27.")
        if any([v for v in spi_connection if v in [26, 27]]):
            log.warning(
                "GPIO pins 26 and 27 are used by USB-Serial/JTAG, "
                "consider using other pins for SPI flash connection."
            )


class ESP32H2StubLoader(StubMixin, ESP32H2ROM):
    """Stub loader for ESP32-H2, runs on top of ROM."""

    pass


ESP32H2ROM.STUB_CLASS = ESP32H2StubLoader



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: GPL-2.0-or-later

from typing import Any

import rich_click as click
from click.parser import OptionParser, ParsingState, _unpack_args

from espefuse.efuse_interface import (
    DEPRECATED_COMMANDS,
    SUPPORTED_BURN_COMMANDS,
    SUPPORTED_COMMANDS,
    SUPPORTED_READ_COMMANDS,
    init_commands,
)
from esptool.cli_util import Group as EsptoolGroup
from esptool.logger import log

click.rich_click.USE_CLICK_SHORT_HELP = True
click.rich_click.COMMAND_GROUPS = {
    "*": [
        {
            "name": "Burn commands",
            "commands": SUPPORTED_BURN_COMMANDS,
        },
        {
            "name": "Read commands",
            "commands": SUPPORTED_READ_COMMANDS,
        },
    ]
}


class ChainParser(OptionParser):
    """
    This is a modified version of the OptionParser class from click.parser.
    It allows for the processing of arguments and options in interspersed order
    together with chaining commands.
    """

    def _process_args_for_options(self, state: ParsingState) -> None:
        while state.rargs:
            arg = state.rargs.pop(0)
            arglen = len(arg)
            # Double dashes always handled explicitly regardless of what
            # prefixes are valid.
            if arg == "--":
                return
            # if the argument is a command, stop parsing options
            elif arg.replace("_", "-") in SUPPORTED_COMMANDS:
                state.largs.append(arg)
                return
            elif arg[:1] in self._opt_prefixes and arglen > 1:
                self._process_opts(arg, state)
            elif self.allow_interspersed_args:
                state.largs.append(arg)
            else:
                state.rargs.insert(0, arg)
                return

    def _process_args_for_args(self, state: ParsingState) -> None:
        pargs, args = _unpack_args(
            state.largs + state.rargs, [x.nargs for x in self._args]
        )

        # This check is required because of the way we modify nargs in ChainingCommand
        if len(pargs) > 0:
            for idx, arg in enumerate(self._args):
                arg.process(pargs[idx], state)

        state.largs = args
        state.rargs = []


class EfuseContext(click.RichContext):
    @property
    def show_sensitive_info(self) -> bool:
        self.ensure_object(dict)
        value: bool = self.obj.get("show_sensitive_info", False)
        if not value:
            log.print("Sensitive data will be hidden (see --show-sensitive-info)")
        return value


class ChainingCommand(click.RichCommand, click.Command):
    context_class = EfuseContext

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _is_option(self, arg: str) -> bool:
        return arg.startswith("--") or arg.startswith("-")

    def invoke(self, ctx: click.Context) -> Any:
        log.print(f'\n=== Run "{self.name}" command ===')
        return super().invoke(ctx)

    def parse_args(self, ctx: click.Context, args: list[str]):
        # This is a hack to set nargs of the last argument to the number of arguments
        # that will be processed separately
        param_changed = None
        for idx, arg in enumerate(args):
            # command found in args or option found after argument
            if arg.replace("_", "-") in SUPPORTED_COMMANDS or (
                self._is_option(arg) and idx > 0
            ):
                arguments_count = sum(
                    isinstance(param, click.Argument) for param in self.params
                )
                for param in self.params:
                    if param.nargs != -1:
                        continue
                    # set nargs of parameter to actual count of arguments and deduct
                    # arguments_count as each argument will be processed separately,
                    # we only care about the last one with nargs=-1
                    # at the end we add 1 to account for the processedargument itself
                    # e.g. if we have burn-bit BLOCK2 1 2 3, we want to set nargs to 3,
                    # so we need to account for BLOCK2 being processed separately
                    param.nargs = args.index(arg) - arguments_count + 1
                    param_changed = param
                    if (
                        param.nargs == 0
                        and param.required
                        and not ctx.resilient_parsing
                    ):
                        raise click.UsageError(
                            f"Command `{self.name}` requires the `{param.name}` "
                            "argument."
                        )
                    break
                break
        ret = super().parse_args(ctx, args)
        # restore nargs of the last argument to -1, in case it is going to be used again
        if param_changed is not None:
            param.nargs = -1
        return ret

    def make_parser(self, ctx: click.Context) -> OptionParser:
        """Creates the underlying option parser for this command."""
        parser = ChainParser(ctx)
        parser.allow_interspersed_args = True
        for param in self.get_params(ctx):
            param.add_to_parser(parser, ctx)
        return parser


class Group(EsptoolGroup):
    DEPRECATED_OPTIONS = {
        "--file_name": "--file-name",
    }

    command_class = ChainingCommand
    context_class = EfuseContext

    @staticmethod
    def _split_to_groups(args: list[str]) -> tuple[list[list[str]], list[str]]:
        """
        This function splits the args list into groups,
        where each item is a cmd with all its args.

        Example:
        all_args:
        ['burn-key-digest', 'secure_images/ecdsa256_secure_boot_signing_key_v2.pem',
         'burn-key', 'BLOCK_KEY0', 'images/efuse/128bit_key',
         'XTS_AES_128_KEY_DERIVED_FROM_128_EFUSE_BITS']

        used_cmds: ['burn-key-digest', 'burn-key']
        groups:
        [['burn-key-digest', 'secure_images/ecdsa256_secure_boot_signing_key_v2.pem'],
         ['burn-key', 'BLOCK_KEY0', 'images/efuse/128bit_key',
          'XTS_AES_128_KEY_DERIVED_FROM_128_EFUSE_BITS']]
        """
        groups: list[list[str]] = []
        args_group: list[str] = []
        used_cmds: list[str] = []
        for arg in args:
            if arg.replace("_", "-") in SUPPORTED_COMMANDS:
                groups.append(args_group)
                used_cmds.append(arg)
                args_group = [arg]
            else:
                args_group.append(arg)
        groups.append(args_group)
        return groups, used_cmds

    @staticmethod
    def repeat_read_commands(
        used_cmds: list[str], groups: list[list[str]]
    ) -> list[list[str]]:
        if (
            sum(cmd in SUPPORTED_BURN_COMMANDS for cmd in used_cmds) > 0
            and sum(cmd in SUPPORTED_READ_COMMANDS for cmd in used_cmds) > 0
        ):
            # append all read commands at the end of group
            read_commands = []
            for group in groups:
                if group[0] in SUPPORTED_READ_COMMANDS:
                    read_commands.append(group)
            groups.extend(read_commands)
        return groups

    def parse_args(self, ctx: click.Context, args: list[str]):
        ctx.ensure_object(dict)
        ctx.obj["is_help"] = any(help_arg in args for help_arg in ctx.help_option_names)
        idx = (
            args.index("--chip")
            if "--chip" in args
            else (args.index("-c") if "-c" in args else -1)
        )
        ctx.obj["chip"] = args[idx + 1] if idx != -1 and idx + 1 < len(args) else "auto"
        # override the default behavior of EsptoolGroup, because we don't need
        # support for parameters with nargs=-1
        args = self._replace_deprecated_args(args)
        cmd_groups, used_cmds = self._split_to_groups(args)

        # Add commands for shell completion
        if ctx.resilient_parsing:
            commands = init_commands(port=None, chip=ctx.obj["chip"], skip_connect=True)
            commands.add_cli_commands(self)
        elif len(used_cmds) == 0:
            log.print(self.get_help(ctx))
            ctx.exit()

        cmd_groups = self.repeat_read_commands(used_cmds, cmd_groups)
        args = [arg for group in cmd_groups for arg in group]

        ctx.obj["used_cmds"] = used_cmds
        ctx.obj["args"] = args
        return super(click.RichGroup, self).parse_args(ctx, args)

    def get_help(self, ctx: click.Context) -> str:
        # help was called without any commands, so we need to add the commands for the
        # default chip
        if not (set(self.list_commands(ctx)) - set(DEPRECATED_COMMANDS)):
            chip = ctx.obj["chip"]
            if chip == "auto":
                log.note(
                    "Chip not specified, showing commands for ESP32 by default. "
                    "Specify the --chip option to get chip-specific help."
                )
                chip = "esp32"
            commands = init_commands(port=None, chip=chip, skip_connect=True)
            commands.add_cli_commands(self)
        return super().get_help(ctx)  # type: ignore



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: GPL-2.0-or-later

from dataclasses import dataclass
from io import StringIO
from typing import Any

import espefuse.efuse.esp32 as esp32_efuse
import espefuse.efuse.esp32c2 as esp32c2_efuse
import espefuse.efuse.esp32c3 as esp32c3_efuse
import espefuse.efuse.esp32c5 as esp32c5_efuse
import espefuse.efuse.esp32c6 as esp32c6_efuse
import espefuse.efuse.esp32c61 as esp32c61_efuse
import espefuse.efuse.esp32e22 as esp32e22_efuse
import espefuse.efuse.esp32h2 as esp32h2_efuse
import espefuse.efuse.esp32h4 as esp32h4_efuse
import espefuse.efuse.esp32h21 as esp32h21_efuse
import espefuse.efuse.esp32p4 as esp32p4_efuse
import espefuse.efuse.esp32s2 as esp32s2_efuse
import espefuse.efuse.esp32s3 as esp32s3_efuse
import espefuse.efuse.esp32s31 as esp32s31_efuse
import esptool
from espefuse.efuse.base_operations import BaseCommands
from espefuse.efuse.emulate_efuse_controller_base import EmulateEfuseControllerBase
from esptool.util import strip_chip_name


@dataclass
class DefChip:
    efuse_lib: Any
    chip_class: type[esptool.ESPLoader]


SUPPORTED_BURN_COMMANDS = [
    "read-protect-efuse",
    "write-protect-efuse",
    "burn-efuse",
    "burn-block-data",
    "burn-bit",
    "burn-key",
    "burn-key-digest",
    "burn-custom-mac",
    "set-flash-voltage",
]

SUPPORTED_READ_COMMANDS = [
    "summary",
    "dump",
    "get-custom-mac",
    "adc-info",
    "check-error",
]

DEPRECATED_COMMANDS = ["execute-scripts"]

SUPPORTED_COMMANDS = (
    SUPPORTED_READ_COMMANDS + SUPPORTED_BURN_COMMANDS + DEPRECATED_COMMANDS
)

SUPPORTED_CHIPS = {
    "esp32": DefChip(esp32_efuse, esptool.targets.ESP32ROM),
    "esp32c2": DefChip(esp32c2_efuse, esptool.targets.ESP32C2ROM),
    "esp32c3": DefChip(esp32c3_efuse, esptool.targets.ESP32C3ROM),
    "esp32c6": DefChip(esp32c6_efuse, esptool.targets.ESP32C6ROM),
    "esp32c61": DefChip(esp32c61_efuse, esptool.targets.ESP32C61ROM),
    "esp32c5": DefChip(esp32c5_efuse, esptool.targets.ESP32C5ROM),
    "esp32e22": DefChip(esp32e22_efuse, esptool.targets.ESP32E22ROM),
    "esp32h2": DefChip(esp32h2_efuse, esptool.targets.ESP32H2ROM),
    "esp32h21": DefChip(esp32h21_efuse, esptool.targets.ESP32H21ROM),
    "esp32h4": DefChip(esp32h4_efuse, esptool.targets.ESP32H4ROM),
    "esp32p4": DefChip(esp32p4_efuse, esptool.targets.ESP32P4ROM),
    "esp32s2": DefChip(esp32s2_efuse, esptool.targets.ESP32S2ROM),
    "esp32s3": DefChip(esp32s3_efuse, esptool.targets.ESP32S3ROM),
    "esp32s31": DefChip(esp32s31_efuse, esptool.targets.ESP32S31ROM),
}


def _get_command_class(chip_name: str) -> BaseCommands:
    return SUPPORTED_CHIPS[chip_name].efuse_lib.commands()  # type: ignore


def init_commands(
    port: str | None = None,
    baud: int = 115200,
    before: str = "default-reset",
    chip: str = "auto",
    esp: esptool.ESPLoader | EmulateEfuseControllerBase | None = None,
    **kwargs: Any,
) -> BaseCommands:
    """Get the ESP eFuse commands class for the given chip
    This function will establish a connection to the chip and
    return the ESP eFuse commands class with initialized chip
    and eFuse values.

    Either esp or port should be provided. If both are provided, esp will be used.
    If neither is provided, the function will create a mock ESPLoader object for tests.

    Args:
        port: The port to connect to the chip
        baud: The baud rate to connect to the chip
        before: The reset mode to use before connecting to the chip
        chip: The chip to use.
        esp: Optional ESPLoader object to use. If provided, the port, baud, before, and
            chip arguments will be ignored. If provided, user has to take care of
            closing the port.

    Keyword Args:
        skip_connect (bool): Whether to skip connecting to the chip. Default is False.
        virt (bool): Whether to use virtual mode. Default is False.
        debug (bool): Whether to enable debug mode. Default is False.
        virt_efuse_file (str): The file to save the eFuse values to. Default is None.
        do_not_confirm (bool): Whether to skip confirmation before burning eFuse.
            Default is False.
        extend_efuse_table (str): The file to extend the eFuse table from.
            Default is None.
        batch_mode (bool): Whether to enable batch mode. Default is False.

    Returns:
        The ESP eFuse commands class
    """
    skip_connect = kwargs.get("skip_connect", False)
    virt = kwargs.get("virt", False)
    debug = kwargs.get("debug", False)
    virt_efuse_file = kwargs.get("virt_efuse_file", None)
    do_not_confirm = kwargs.get("do_not_confirm", False)
    extend_efuse_table = kwargs.get("extend_efuse_table", None)
    external_esp = esp is not None
    batch_mode = kwargs.get("batch_mode", False)

    if esp is None:
        esp = get_esp(
            port, baud, before, chip, skip_connect, virt, debug, virt_efuse_file
        )

    try:
        commands = _get_command_class(strip_chip_name(esp.CHIP_NAME))
        commands.esp = esp
        commands.external_esp = external_esp
        commands.get_efuses(
            skip_connect=skip_connect,
            debug_mode=debug,
            do_not_confirm=do_not_confirm,
            extend_efuse_table=extend_efuse_table,
        )
    except Exception:
        # If creating commands fails, ensure the port is closed
        BaseCommands._close_port(esp, external_esp)
        raise
    if batch_mode:
        commands.use_batch_mode()
    return commands


def get_esp(
    port: str | None = None,
    baud: int = 115200,
    before: str = "default-reset",
    chip: str = "auto",
    skip_connect: bool = False,
    virt: bool = False,
    debug: bool = False,
    virt_efuse_file: str | None = None,
) -> esptool.ESPLoader | EmulateEfuseControllerBase:
    """Get the ESPLoader object for the given chip.
    Uses :func:`esptool.cmds.detect_chip` function.

    Args:
        port: The port to connect to the chip
        baud: The baud rate to connect to the chip
        before: The reset mode to use before connecting to the chip
            Supported values are: "default-reset", "usb-reset", "no-reset",
            "no-reset-no-sync"
        chip: The chip to use
        skip_connect: Whether to skip connecting to the chip
        virt: Whether to use virtual mode
        debug: Whether to enable debug mode
        virt_efuse_file: The file to save the eFuse values to

    Returns:
        The ESPLoader object or EmulateEfuseController object
    """
    if chip not in ["auto"] + list(SUPPORTED_CHIPS.keys()):
        raise esptool.FatalError(f"get_esp: Unsupported chip ({chip})")

    if virt:
        efuse = SUPPORTED_CHIPS.get(chip, SUPPORTED_CHIPS["esp32"]).efuse_lib
        return efuse.EmulateEfuseController(virt_efuse_file, debug)  # type: ignore

    if chip == "auto" and not skip_connect:
        if port is None:
            raise esptool.FatalError(
                "get_esp: Port is required when chip is 'auto' to detect the chip"
            )
        return esptool.detect_chip(port, baud, before)

    esp = SUPPORTED_CHIPS.get(chip, SUPPORTED_CHIPS["esp32"]).chip_class(
        port if not skip_connect else StringIO(),  # type: ignore
        baud,
    )
    if not skip_connect:
        esp.connect(before)
        if esp.sync_stub_detected:
            esp = esp.STUB_CLASS(esp)  # type: ignore
    return esp



# SPDX-FileCopyrightText: 2014-2025 Fredrik Ahlberg, Angus Gratton,
# Espressif Systems (Shanghai) CO LTD, other contributors as noted.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import annotations

import os
import re
import struct
from typing import IO, TypeAlias

# Define a custom type for the input
ImageSource: TypeAlias = str | bytes | IO[bytes]


def byte(bitstr, index):
    return bitstr[index]


def mask_to_shift(mask):
    """Return the index of the least significant bit in the mask"""
    shift = 0
    while mask & 0x1 == 0:
        shift += 1
        mask >>= 1
    return shift


def div_roundup(a, b):
    """Return a/b rounded up to nearest integer,
    equivalent result to int(math.ceil(float(int(a)) / float(int(b))), only
    without possible floating point accuracy errors.
    """
    return (int(a) + int(b) - 1) // int(b)


def flash_size_bytes(size):
    """Given a flash size of the type passed in size
    (ie 512KB or 1MB) then return the size in bytes.
    """
    if size is None:
        return None
    if "MB" in size:
        return int(size[: size.index("MB")]) * 1024 * 1024
    elif "KB" in size:
        return int(size[: size.index("KB")]) * 1024
    else:
        raise FatalError(f"Unknown size {size}")


def hexify(s, uppercase=True):
    format_str = "%02X" if uppercase else "%02x"
    return "".join(format_str % c for c in s)


def pad_to(data, alignment, pad_character=b"\xff"):
    """Pad to the next alignment boundary"""
    pad_mod = len(data) % alignment
    if pad_mod != 0:
        data += pad_character * (alignment - pad_mod)
    return data


def expand_chip_name(chip_name):
    """Change chip name to official form, e.g. `esp32s3beta2` -> `ESP32-S3(beta2)`"""
    # Put "-" after "esp32"
    chip_name = re.sub(r"(esp32)(?!$)", r"\1-", chip_name)
    # Put "()" around "betaN"
    chip_name = re.sub(r"(beta\d*)", r"(\1)", chip_name)
    # Uppercase everything before "(betaN)"
    chip_name = re.sub(r"^[^\(]+", lambda x: x.group(0).upper(), chip_name)
    return chip_name


def strip_chip_name(chip_name):
    """Strip chip name to normalized form, e.g. `ESP32-S3(beta2)` -> `esp32s3beta2`"""
    return re.sub(r"[-()]", "", chip_name.lower())


def get_file_size(path_to_file):
    """Returns the file size in bytes"""
    file_size = 0
    with open(path_to_file, "rb") as f:
        f.seek(0, os.SEEK_END)
        file_size = f.tell()
    return file_size


def sanitize_string(byte_string):
    return byte_string.decode("utf-8").replace("\0", "")


def get_bytes(input: ImageSource) -> tuple[bytes, str | None]:
    """
    Normalize the input (file path, bytes, or an opened file-like object) into bytes
    and provide a name of the source.

    Args:
        input: The input file path, bytes, or an opened file-like object.

    Returns:
        A tuple containing the normalized bytes and the source of the input.
    """
    if isinstance(input, str):
        with open(input, "rb") as f:
            data = f.read()
            source = input
    elif isinstance(input, bytes):
        data = input
        source = None
    elif hasattr(input, "read") and hasattr(input, "write") and hasattr(input, "close"):
        pos = input.tell()
        data = input.read()
        input.seek(pos)  # Reset the file pointer
        source = input.name
    else:
        raise FatalError(f"Invalid input type {type(input)}")
    return data, source


def get_key_from_value(dict, val):
    """
    Get key from value in dictionary, assumes unique values in dictionary
    """
    for key, value in dict.items():
        if value == val:
            return key
    return None


def check_deprecated_py_suffix(module_name: str) -> None:
    """Check if called with deprecated .py suffix"""
    import sys

    from esptool import log

    script_name = sys.argv[0] if sys.argv else ""
    if script_name.endswith(module_name + ".py"):
        log.warning(
            f"DEPRECATED: '{module_name}.py' is deprecated. Please use '{module_name}' "
            "instead. The '.py' suffix will be removed in a future major release."
        )


class PrintOnce:
    """
    Class for printing messages just once. Can be useful when running in a loop
    """

    def __init__(self, print_callback) -> None:
        self.already_printed = False
        self.print_callback = print_callback

    def __call__(self, text) -> None:
        if not self.already_printed:
            self.print_callback(text)
            self.already_printed = True


class FatalError(RuntimeError):
    """
    Wrapper class for runtime errors that aren't caused by internal bugs, but by
    ESP ROM responses or input content.
    """

    def __init__(self, message):
        RuntimeError.__init__(self, message)

    @staticmethod
    def WithResult(message, result):
        """
        Return a fatal error object that appends the hex values of
        'result' and its meaning as a string formatted argument.
        """

        err_defs = {
            # ROM error codes
            0x100: "Undefined errors",
            0x101: "The input parameter is invalid",
            0x102: "Failed to malloc memory from system",
            0x103: "Failed to send out message",
            0x104: "Failed to receive message",
            0x105: "The format of the received message is invalid",
            0x106: "Message is ok, but the running result is wrong",
            0x107: "Checksum error",
            0x108: "Flash write error",
            0x109: "Flash read error",
            0x10A: "Flash read length error",
            0x10B: "Deflate failed error",
            0x10C: "Deflate Adler32 error",
            0x10D: "Deflate parameter error",
            0x10E: "Invalid RAM binary size",
            0x10F: "Invalid RAM binary address",
            0x164: "Invalid parameter",
            0x165: "Invalid format",
            0x166: "Description too long",
            0x167: "Bad encoding description",
            0x169: "Insufficient storage",
            # Flasher stub error codes
            0xC000: "Bad data length",
            0xC100: "Bad data checksum",
            0xC200: "Bad blocksize",
            0xC300: "Invalid command",
            0xC400: "Failed SPI operation",
            0xC500: "Failed SPI unlock",
            0xC600: "Not in flash mode",
            0xC700: "Inflate error",
            0xC800: "Not enough data",
            0xC900: "Too much data",
            0xCA00: "NAND program failed (P_FAIL)",
            0xCB00: "NAND erase failed (E_FAIL)",
            0xFF00: "Command not implemented",
        }

        err_code = struct.unpack(">H", result[:2])
        message += " (result was {}: {})".format(
            hexify(result), err_defs.get(err_code[0], "Unknown result")
        )
        if err_code[0] == 0xCA00:
            return NANDProgramFailed(message)
        if err_code[0] == 0xCB00:
            return NANDEraseFailed(message)
        return FatalError(message)


class NotImplementedInROMError(FatalError):
    """
    Wrapper class for the error thrown when a particular ESP bootloader function
    is not implemented in the ROM bootloader.
    """

    def __init__(self, bootloader, func):
        FatalError.__init__(
            self,
            f"{bootloader.CHIP_NAME} ROM does not support function {func.__name__}.",
        )


class NotSupportedError(FatalError):
    def __init__(self, esp, function_name):
        FatalError.__init__(
            self,
            f"{function_name} is not supported by {esp.CHIP_NAME}.",
        )


class NANDProgramFailed(FatalError):
    """Raised when the stub reports a NAND P_FAIL (program failed, 0xCA00).

    Indicates the chip set the P_FAIL bit after a page program operation.
    Per W25N01GV app-note: mark the block bad and retry on the next block.
    """

    def __init__(self, message):
        FatalError.__init__(self, message)


class NANDEraseFailed(FatalError):
    """Raised when the stub reports a NAND E_FAIL (erase failed, 0xCB00).

    Indicates the chip set the E_FAIL bit after a block erase operation.
    Per W25N01GV app-note: mark the block bad and retry on the next block.
    """

    def __init__(self, message):
        FatalError.__init__(self, message)


class UnsupportedCommandError(RuntimeError):
    """
    Wrapper class for when ROM loader returns an invalid command response.

    Usually this indicates the loader is running in Secure Download Mode.
    """

    def __init__(self, esp, op):
        if esp.secure_download_mode:
            msg = f"This command ({op:#x}) is not supported in Secure Download Mode"
        else:
            msg = f"Invalid (unsupported) command {op:#x}"
        RuntimeError.__init__(self, msg)
